# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import sys
import logging
from bunch import Bunch
from docopt import docopt
from kotori.version import __VERSION__
from kotori.util.common import setup_logging
from kotori.util.configuration import get_configuration, get_configuration_file
from kotori.vendor.lst.application import setup_binary_message_adapter
from kotori.vendor.lst.commands import lst_channels, lst_message, sanitize_channel_label, compute_channel_label

logger = logging.getLogger(__name__)

APP_NAME = 'lst-message ' + __VERSION__

def message():
    """
    Usage:
      lst-message list-channels                    [--config etc/kotori.ini] [--debug]
      lst-message <channel>  decode     <payload>  [--config etc/kotori.ini] [--debug]
      lst-message <channel>  transform  <payload>  [--config etc/kotori.ini] [--debug]
      lst-message <channel>  send       <payload>  --target=udp://localhost:8888 [--config etc/kotori.ini]
      lst-message <channel>  info       [<name>]   [--config etc/kotori.ini] [--debug]
      lst-message --version
      lst-message (-h | --help)

    Options:
      --config etc/kotori.ini   Use specified configuration file, otherwise try KOTORI_CONFIG environment variable
      --version                 Show version information
      --debug                   Enable debug messages
      -h --help                 Show this screen

    Examples:

      # Initialize environment with path to configuration file, either on development or production
      export KOTORI_CONFIG=`pwd`/etc/development.ini
      export KOTORI_CONFIG=/etc/kotori/kotori.ini

      # Show list of channels
      lst-message list-channels

      # Display list of structs for channel "h2m"
      lst-message h2m info

      # Display details about struct "struct_fuelcell_r" in channel "h2m"
      lst-message h2m info struct_fuelcell_r

      # Decode message for channel "h2m"
      lst-message h2m decode 0x05022a0021

      # Send message to channel "h2m"
      lst-message h2m send 0x05022a0021 --target=udp://localhost:8888

    """

    # parse command line arguments
    options = docopt(message.__doc__, version=APP_NAME)
    #print 'options: {}'.format(options)

    debug = options.get('--debug')

    # setup logging
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    setup_logging(log_level)

    # load configuration
    config_file = get_configuration_file(options.get('--config'))
    config = get_configuration(config_file)

    # first command
    if options.get('list-channels'):
        lst_channels(config)
        sys.exit()

    # activate/mount application representing selected channel
    channel_label = options.get('<channel>')
    channel_name  = sanitize_channel_label(channel_label)
    if channel_name == channel_label:
        channel_label = compute_channel_label(channel_name)
    channel = Bunch(
        name  = channel_name,
        label = channel_label)
    # TODO: overhaul _active_ mechanics
    try:
        config['_active_'] = config[channel.label]
        channel.settings = config['_active_']
    except KeyError:
        logger.error('Channel configuration object "{channel_label}" not found'.format(channel_label = channel.label))
        raise

    # initialize library, setup BinaryMessageAdapter
    adapter = setup_binary_message_adapter(config)

    # dispatch to generic message decoding
    lst_message(channel, adapter, options)



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl, <andreas@getkotori.org>
import re
from kotori.util.common import SmartBunch


class WanBusStrategy(object):

    def __init__(self, *args, **kwargs):
        pass

    def topic_to_topology(self, topic):
        """
        Decode MQTT topic segments implementing the »quadruple hierarchy strategy«.

        The topology hierarchy is directly specified by the MQTT topic and is
        made up of a minimum of four identifiers describing the core structure::

            realm / network / gateway / node

        The topology identifiers are specified as:

            - "realm" is the designated root realm. You should prefix the topic name
              with this label when opting in for all features of the telemetry platform.
              For other purposes, feel free to publish to any MQTT topic you like.

            - "network" is your personal realm. Choose anything you like or use an
              `Online GUID Generator <https://www.guidgenerator.com/>`_ to gain
              maximum uniqueness.

            - "gateway" is your gateway identifier. Choose anything you like.
              This does not have to be very unique, so you might use labels
              having the names of sites. While you are the owner of this
              namespace hierarchy, remember these labels might be visible on
              the collaborative ether, though.
              So the best thing would be to give kind of nicknames to your
              sites which don't identify their location.

            - "1" is your node identifier. Choose anything you like. This usually
              gets transmitted from an embedded device node. Remember one device node
              might have multiple sensors attached, which is beyond the scope of the
              collection platform: We just accept bunches of named measurement values,
              no matter which sensors they might originate from.
              In other words: We don't need nor favor numeric sensor identifiers,
              let's give them names!
        """

        # regular expression pattern for decoding MQTT topic address segments
        pattern = r'^(?P<realm>.+?)/(?P<network>.+?)/(?P<gateway>.+?)/(?P<node>.+?)(?:/(?P<slot>.+?))?$'

        # decode the topic
        p = re.compile(pattern)
        m = p.match(topic)
        if m:
            address = SmartBunch(m.groupdict())
        else:
            address = {}

        return address


    @classmethod
    def topology_to_storage(self, topology):
        """
        Encode topology segment identifiers to database address.

        A database server usually has the concept of multiple databases,
        each with multiple tables. With other databases than RDBMS,
        they might be named differently, but the concept in general
        is the same.

        When mapping the topology quadruple (realm, network, gateway, node) in the form of:

            - realm + network = database name
            - gateway + node  = table name

        We have a perfect fit for computing the slot where to store the measurements.

        """

        # Todo: Investigate using tags additionally to / instead of database.measurement
        # Todo: Move specific stuff about WeeWX or Tasmota to some device-specific knowledgebase.

        # data:     Regular endpoint
        # loop:     WeeWX
        # SENSOR:   Sonoff-Tasmota
        if topology.slot.startswith('data') or topology.slot.startswith('loop') \
                or topology.slot.endswith('SENSOR') or topology.slot.endswith('STATE'):
            suffix = 'sensors'

        elif topology.slot.startswith('event'):
            suffix = 'events'

        else:
            suffix = 'unknown'

        # Use topology information as blueprint for storage address
        storage = SmartBunch(topology)

        # Format and sanitize all input parameters used for database addressing
        sanitize = self.sanitize_db_identifier
        storage.label       = sanitize('{}-{}'.format(storage.gateway, storage.node))
        storage.database    = sanitize('{}_{}'.format(storage.realm, storage.network))
        storage.measurement = sanitize('{}_{}_{}'.format(storage.gateway, storage.node, suffix))
        storage.measurement_events = sanitize('{}_{}_{}'.format(storage.gateway, storage.node, 'events'))

        return storage

    @classmethod
    def topology_to_label(cls, topology):
        return '{}-{}-{}-{}'.format(topology.realm, topology.network, topology.gateway, topology.node)

    @staticmethod
    def sanitize_db_identifier(value):
        """
        Different databases accept different special
        characters as database- or table names.

        Better safe than sorry, let's strip them all.
        """
        value = value.replace('/', '_').replace('.', '_').replace('-', '_').lower()
        return value



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
import logging
from pkg_resources import resource_filename
from pyramid.httpexceptions import HTTPFound, HTTPNotFound
from pyramid.response import Response, FileResponse
from pyramid.url import route_path
from pyramid.view import view_config

log = logging.getLogger(__name__)

def includeme(config):

    # serve favicon.ico
    # http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/assets.html#registering-a-view-callable-to-serve-a-static-asset
    config.add_route('favicon', 'favicon.ico')
    config.add_view('kotori.frontend.views.favicon_view', route_name='favicon')


@view_config(route_name='index', renderer='kotori.frontend:templates/index.html')
def index_page(request):
    software_version = request.registry.settings['SOFTWARE_VERSION']
    tplvars = {
        'software_version': software_version,
    }
    return tplvars


def favicon_view(request):
    """
    http://docs.pylonsproject.org/projects/pyramid/en/latest/narr/assets.html#registering-a-view-callable-to-serve-a-static-asset
    """
    icon = resource_filename('kotori.frontend', 'static/favicon.ico')
    if os.path.isfile(icon):
        return FileResponse(icon, request=request)
    else:
        return HTTPNotFound()



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from twisted.logger import Logger
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import DatagramProtocol
from autobahn.twisted.wamp import ApplicationSession
from kotori.bus.wamp import WampBus
from kotori.vendor.hydro2motion.util.geo import turn_xyz_into_llh

log = Logger()

app_session = None

class UdpPublisher(ApplicationSession):
    """
    Simple message publishing with Autobahn WebSockets.

    derived from:
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/basic/pubsub/basic/frontend.py
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/pubsub/simple/example2/client.py
    - https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/session/fromoutside/client.py
    """

    def onJoin(self, details):
        log.info("UdpPublisher joined WAMP bus")

        global app_session
        app_session = self


    def onDisconnect(self):
        log.info("UdpPublisher disconnected from WAMP bus")



class UdpAdapter(DatagramProtocol):
    # https://twistedmatrix.com/documents/15.0.0/core/howto/udp.html

    @inlineCallbacks
    def datagramReceived(self, data, (host, port)):
        log.info("Received via UDP from %s:%d: %r " % (host, port, data))

        try:
            payload = data.split(';')
            GPS_X           = int(float(payload[24]))
            GPS_Y           = int(float(payload[25]))
            GPS_Z           = int(float(payload[26]))

            x = GPS_X / 100.0
            y = GPS_Y / 100.0
            z = GPS_Z / 100.0


            llh = turn_xyz_into_llh(x, y, z, "wgs84")
            lat = llh[0]
            lng = llh[1]

            seq = (data, str(lat), str(lng))

            jdata = ";".join(seq)

        except ValueError:
            log.warn('Could not decode data: {}'.format(data))

        # ECHO
        #self.transport.write(data, (host, port))

        # forward
        yield app_session.publish(u'de.elmyra.kotori.telemetry.data', jdata)


def listen_udp(udp_port):
    log.info('Starting udp adapter on port "{}"'.format(udp_port))
    reactor.listenUDP(udp_port, UdpAdapter())


def h2m_boot_udp_adapter(settings, debug=False):

    websocket_uri = unicode(settings.wamp.uri)
    udp_port = int(settings.hydro2motion.udp_port)

    wamp_bus = WampBus(uri=websocket_uri, session_factory=UdpPublisher)
    wamp_bus.connect()
    listen_udp(udp_port)



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from kotori.version import __VERSION__
from pyramid.config import Configurator

def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""

    settings['SOFTWARE_VERSION'] = __VERSION__

    config = Configurator(settings=settings)

    # Addons
    config.include('pyramid_jinja2')
    # http://docs.pylonsproject.org/projects/pyramid-jinja2/en/latest/#adding-or-overriding-a-renderer
    config.add_jinja2_renderer('.html')
    config.include('cornice')

    # Views and routes
    config.add_static_view('static/app', 'static/app', cache_max_age=0)
    config.add_static_view('static/lib', 'static/lib', cache_max_age=60 * 24)
    config.add_route('index', '/')

    config.scan()

    return config.make_wsgi_app()



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl <andreas@getkotori.org>
import time
import json

import arrow
from bunch import Bunch
from cornice.util import to_list
from twisted.logger import Logger, LogLevel
from twisted.internet import reactor, threads
from twisted.internet.task import LoopingCall
from twisted.application.service import MultiService, Service
from twisted.python.failure import Failure
from twisted.python.threadpool import ThreadPool

from kotori.daq.decoder.tasmota import TasmotaDecoder
from kotori.daq.services.schema import MessageType, TopicMatchers
from kotori.daq.services import MultiServiceMixin
from kotori.daq.intercom.mqtt import MqttAdapter
from kotori.daq.storage.influx import InfluxDBAdapter
from kotori.io.protocol.util import convert_floats
from kotori.util.configuration import read_list
from kotori.util.thimble import Thimble

log = Logger()


class MqttInfluxGrafanaService(MultiService, MultiServiceMixin):

    def __init__(self, channel=None, graphing=None, strategy=None):

        MultiService.__init__(self)

        # TODO: Sanity checks/assertions against channel, graphing and strategy

        # TODO: Make subsystems dynamic
        self.subsystems = ['channel', 'graphing', 'strategy']
        self.channel = channel or Bunch(realm=None, subscriptions=[])
        self.graphing = to_list(graphing)
        self.strategy = strategy

        self.name = u'service-mig-' + self.channel.get('realm', unicode(id(self)))

    def setupService(self):

        self.log(log.info, u'Bootstrapping')
        self.settings = self.parent.settings

        # Optionally register subsystem component as child service
        for subsystem in self.subsystems:
            if hasattr(self, subsystem):
                subsystem_service = getattr(self, subsystem)
                if isinstance(subsystem_service, Service):
                    log.info('Registering subsystem component "{subsystem}" as service', subsystem=subsystem)
                    self.registerService(subsystem_service)

        # Configure metrics to be collected each X seconds
        metrics_interval = int(self.channel.get('metrics_logger_interval', 60))
        self.metrics = Bunch(tx_count=0, starttime=time.time(), interval=metrics_interval)

        subscriptions = read_list(self.channel.mqtt_topics)
        self.mqtt_service = MqttAdapter(
            name          = u'mqtt-' + self.channel.realm,
            broker_host   = self.settings.mqtt.host,
            broker_port   = int(self.settings.mqtt.port),
            broker_username = self.settings.mqtt.username,
            broker_password = self.settings.mqtt.password,
            callback      = self.mqtt_receive,
            subscriptions = subscriptions)

        self.registerService(self.mqtt_service)

        self.influx = InfluxDBAdapter(settings = self.settings.influxdb)

        # Perform MQTT message processing using a different thread pool
        self.threadpool = ThreadPool()
        self.thimble = Thimble(reactor, self.threadpool, self, ["process_message"])

    def startService(self):
        self.setupService()
        #self.log(log.info, u'Starting')
        MultiService.startService(self)
        self.metrics_twingo = LoopingCall(self.process_metrics)
        self.metrics_twingo.start(self.metrics.interval, now=True)

    def log(self, level, prefix):
        level(u'{prefix} {class_name}. name={name}, channel={channel}',
            prefix=prefix, class_name=self.__class__.__name__, name=self.name, channel=dict(self.channel))

    def topic_to_topology(self, topic):
        return self.strategy.topic_to_topology(topic)

    def topology_to_storage(self, topology):
        return self.strategy.topology_to_storage(topology)

    def get_basetopic(self, topic):
        topic = TopicMatchers.data.sub('', topic)
        topic = TopicMatchers.event.sub('', topic)
        return topic

    def classify_topic(self, topic):
        if TopicMatchers.data.search(topic): return MessageType.DATA_CONTAINER
        if TopicMatchers.discrete.search(topic): return MessageType.DATA_DISCRETE
        if TopicMatchers.event.search(topic): return MessageType.EVENT
        if TopicMatchers.error.search(topic): return MessageType.ERROR

    def mqtt_receive(self, topic=None, payload=None, **kwargs):
        try:
            # Synchronous message processing
            #return self.process_message(topic, payload, **kwargs)

            # Asynchronous message processing
            #deferred = threads.deferToThread(self.process_message, topic, payload, **kwargs)

            # Asynchronous message processing using different thread pool
            deferred = self.thimble.process_message(topic, payload, **kwargs)

            deferred.addErrback(self.mqtt_process_error, topic, payload)
            deferred.addErrback(self.mqtt_exception, topic, payload)
            return deferred

        except Exception:
            log.failure(u'Processing MQTT message failed. topic={topic}, payload={payload}', topic=topic, payload=payload)

    def process_message(self, topic, payload, **kwargs):

        payload = payload.decode('utf-8')

        # Ignore MQTT error signalling messages
        if topic.endswith('error.json'):
            return

        if self.channel.realm and not topic.startswith(self.channel.realm):
            #log.info('Ignoring message to topic {topic}, realm={realm}', topic=topic, realm=self.channel.realm)
            return False

        log.debug(u"Processing message on topic '{topic}' with payload '{payload}'", topic=topic, payload=payload)

        # Compute storage address from topic
        topology = self.topic_to_topology(topic)
        log.debug(u'Topology address: {topology}', topology=dict(topology))

        message_type = self.classify_topic(topic)
        message = None

        # a) En bloc: Multiple measurements in JSON object
        #
        # The suffixes are:
        #
        #   - data.json:        Regular
        #   - data/__json__:    Homie
        #   - loop:             WeeWX
        #   - message-json:     Deprecated
        #
        # Todo: Move specific stuff about Homie, WeeWX or Tasmota to some device-specific knowledgebase.

        if message_type == MessageType.DATA_CONTAINER:

            # Decode regular JSON container.
            message = json.loads(payload)

            # Required for WeeWX data
            #message = convert_floats(json.loads(payload))

            decoders = [
                # Decoder for Sonoff-Tasmota telemetry payload.
                TasmotaDecoder,
            ]

            for decoder_class in decoders:
                decoder = decoder_class(topology=topology)
                message = decoder.decode_message(message)

        # b) Discrete values
        #
        # The suffixes are:
        #
        #   - data/temperature
        #   - data/humidity
        #   - ...
        #
        elif message_type == MessageType.DATA_DISCRETE:

            # Todo: Backward compat for single readings - refactor elsewhere.
            if 'slot' in topology and topology.slot.startswith('measure/'):
                topology.slot = topology.slot.replace('measure/', 'data/')

            # Single measurement as plain value; assume float
            # Convert to MessageType.DATA_CONTAINER
            if 'slot' in topology and topology.slot.startswith('data/'):

                # This is sensor data
                message_type = MessageType.DATA_CONTAINER

                # Amend topic and compute storage message from single scalar value
                name = topology.slot.replace('data/', '')
                value = float(payload)
                message = {name: value}

        # Set an event
        elif message_type == MessageType.EVENT:

            # This is an event
            message_type = MessageType.EVENT

            # Decode message from json format
            message = json.loads(payload)

        # Catch an error message
        # TODO: Signal via MQTT
        elif message_type == MessageType.ERROR:
            log.debug(u'Ignoring error message from MQTT, "{topic}" with payload "{payload}"', topic=topic, payload=payload)
            return

        else:
            log.debug(u'Unknown message type on topic "{topic}" with payload "{payload}", ignoring.', topic=topic, payload=payload)
            return

        # count transaction
        self.metrics.tx_count += 1

        # TODO: Re-enable for measuring packet ingress frequency.
        # Currently turned off since sending timestamps from data acquisition.
        """
        if 'time' in message:
            self.metrics.packet_time = message['time']
        else:
            self.metrics.packet_time = None
        """

        # TODO: Data enrichment machinery, e.g. for geospatial data
        # latitude/lat, longitude/long/lon/lng
        # 1. geohash => lat/lon
        # 2. postcode/city => lat/lon (forward geocoder)
        # 3. lat/lon => geohash if not geohash
        # 4. lat/lon => reverse geocoder (address information)
        # The outcome can provide additional meta information to be used by the tagging machinery below,
        # e.g. create tags from homogenized Nomatim address modulo "house_number" etc.

        # TODO: Already do the tagging enrichment machinery here(!) to
        # establish additional metadata schema for further processing, e.g. Grafana.
        # So, move the schwumms from storage handler here!
        # Sane order for Grafana template variables:
        # continent, country_code (upper), q-region, city, q-hood, road, (compound)

        # Compute storage location
        storage_location = self.topology_to_storage(topology)
        log.debug(u'Storage location: {storage}', storage=dict(storage_location))

        # Store data or event
        if message_type in (MessageType.DATA_CONTAINER, MessageType.EVENT):
            self.store_message(storage_location, message)

        # Provision graphing subsystem
        if message_type == MessageType.DATA_CONTAINER:
            # TODO: Purge message from fields to be used as tags
            # Namely:
            # 'geohash',
            # 'location', 'location_id', 'location_name', 'sensor_id', 'sensor_type',
            # 'latitude', 'longitude', 'lat', 'lon'
            for graphing_subsystem in self.graphing:

                # Mix in references to each other. A bit of a hack, but okay for now :-).
                graphing_subsystem.strategy = self.strategy

                subsystem_name = graphing_subsystem.__class__.__name__
                log.debug(u'Provisioning Grafana with {name}', name=subsystem_name)
                try:
                    graphing_subsystem.provision(storage_location, message, topology=topology)

                except Exception as ex:
                    log.failure(u'Grafana provisioning failed for storage={storage}, message={message}:\n{log_failure}',
                                storage=storage_location.dump(), message=message,
                                level=LogLevel.error)

                    # MQTT error signalling
                    failure = Failure()
                    self.mqtt_publish_error(failure, topic, payload)

        return True

    def store_message(self, storage, data):
        """
        Store data to timeseries database

        :param storage: The storage location object
        :param data:    The data ready for storing
        """
        self.influx.write(storage, data)

    def mqtt_process_error(self, failure, topic, payload):
        """
        Failure handling

        :param failure: Failure object from Twisted
        :param topic:   Full MQTT topic
        :param payload: Raw MQTT payload
        """

        # Log failure
        log.failure(u'Processing MQTT message failed from topic "{topic}":\n{log_failure}', topic=topic, failure=failure, level=LogLevel.error)

        # MQTT error signalling
        self.mqtt_publish_error(failure, topic, payload)

    def mqtt_exception(self, failure, topic, payload):
        log.failure(u'Problem publishing error message:\n{log_failure}', failure=failure, level=LogLevel.warn)

    def mqtt_publish_error(self, failure, topic, payload):
        """
        Error signalling over MQTT to "error.json" topic suffix

        :param failure: Failure object from Twisted
        :param topic:   Full MQTT topic
        :param payload: Raw MQTT payload
        """

        # Compute base topic of data acquisition channel
        basetopic = self.get_basetopic(topic)
        log.debug(u'Channel base topic is {basetopic}', basetopic=basetopic)

        # Effective error reporting topic
        error_topic = basetopic + '/' + 'error.json'

        #
        error = {
            'type': unicode(failure.type),
            'message': failure.getErrorMessage(),
            'description': u'Error processing MQTT message "{payload}" from topic "{topic}".'.format(topic=topic, payload=payload),
            'timestamp': arrow.utcnow().format('YYYY-MM-DDTHH:mm:ssZZ'),
            #'failure': unicode(failure),
        }
        message = json.dumps(error, indent=4)

        # Publish error signal over MQTT
        #log.debug('Publishing error message to topic {topic}: {message}', topic=error_topic, message=message)
        self.mqtt_service.publish(error_topic, message)

    def process_metrics(self):

        metrics = []

        # Compute frequency of measurements
        if 'packet_time' in self.metrics and self.metrics['packet_time'] is not None:

            self.metrics.setdefault('packet_starttime', self.metrics.packet_time)

            # Convert nanos to seconds
            packet_duration = (self.metrics.packet_time - self.metrics.packet_starttime) / 1000.0 / 1000.0 / 1000.0
            packet_duration = packet_duration or self.metrics.starttime
            if packet_duration != 0:
                packet_frequency = self.metrics.tx_count / float(packet_duration)
            else:
                packet_frequency = 0.0

            metrics.append('measurements: %.02f Hz' % packet_frequency)

            # Reset for next round
            self.metrics.packet_starttime = self.metrics.packet_time

        # Compute frequency of transactions
        now = time.time()
        transaction_duration = now - self.metrics.starttime
        if transaction_duration != 0:
            transaction_frequency = self.metrics.tx_count / float(transaction_duration)
        else:
            transaction_frequency = 0.0
        metrics.append('transactions: %.02f tps' % transaction_frequency)

        # Reset for next round
        self.metrics.tx_count = 0
        self.metrics.starttime = now


        # Add information from the Twisted reactor
        pending_calls = reactor.getDelayedCalls()
        pending_count = len(pending_calls)
        #metrics.append('pending: %d' % pending_count)

        metrics_info = ', '.join(metrics)

        log.info('[{realm:12s}] {metrics_info}', realm=self.channel.realm, metrics_info=metrics_info)



# -*- coding: utf-8 -*-
# (c) 2014-2015 Andreas Motl, <andreas@getkotori.org>
import sys
import traceback
from StringIO import StringIO

def traceback_get_exception(num = -1):

    # build error message
    exception_string = ''.join(traceback.format_exception_only(sys.exc_type, hasattr(sys, 'exc_value') and sys.exc_value or 'Unknown'))

    # extract error location from traceback
    if hasattr(sys, 'exc_traceback'):
        (filename, line_number, function_name, text) = traceback.extract_tb(sys.exc_traceback)[num]
    else:
        (filename, line_number, function_name, text) = ('-', '-', '-', '-')

    error = {
        'message': exception_string,
        'location': {
            'filename': filename,
            'line_number': line_number,
            'function_name': function_name,
            'text': text,
            }
    }

    return error

def format_exception_location(error, prefix=''):
    if prefix:
        prefix += "\n"
    error_location =\
    prefix +\
    "Filename:    %s\nLine number: %s\nFunction:    %s\nCode:        %s" %\
    (error['location']['filename'], error['location']['line_number'], error['location']['function_name'], error['location']['text'])
    return error_location


def last_error_and_traceback():
    error_entry = traceback_get_exception(0)
    error_last = traceback_get_exception(-1)
    sep = '-' * 60
    payload_location = "\n".join([sep, format_exception_location(error_entry, "Entry point:"), sep, format_exception_location(error_last, "Source of exception:")])
    payload = ''.join(["ERROR: ", error_entry['message'], "\n", payload_location, "\n"])

    # add full traceback
    buffer = StringIO()
    traceback.print_exc(file=buffer)
    buffer.seek(0)
    payload += '\n' + buffer.read()

    return payload




# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
import math
import arrow
import types
import datetime
from six import text_type
from dateutil.tz import gettz
from dateutil.parser import parse
from pyramid.settings import asbool
from twisted.web import http
from twisted.logger import Logger
from twisted.python.url import URL
from twisted.python.failure import Failure
from twisted.web.error import Error

log = Logger()

def get_data_uri(bucket, sibling=None, more_params=None):
    """
    Compute uri to data source as sibling to the current path.
    Add "from" and "to" query parameters from bucket.
    """

    more_params = more_params or {}

    forward_parameters = [u'from', u'to', u'exclude', u'include', u'pad', u'backfill', u'interpolate']

    request = bucket.request

    # Honor X-Forwarded-Proto request header if behind SSL-terminating HTTP proxy
    twisted_honor_reverse_proxy(request)

    url = URL()
    for param in forward_parameters:
        if param in bucket.tdata:
            url = url.add(unicode(param), unicode(bucket.tdata[param]))

    for param, value in more_params.iteritems():

        # Special rule: Don't add any of "pad" or "backfill", if "interpolate" is true
        do_interpolate = 'interpolate' in bucket.tdata and asbool(bucket.tdata.interpolate)
        if do_interpolate and param in ['pad', 'backfill']:
            continue

        url = url.add(unicode(param), unicode(value))

    data_uri = str(request.URLPath().sibling(sibling).click(url.asText()))
    return data_uri

def twisted_honor_reverse_proxy(request):
    # Honor X-Forwarded-Proto request header if behind SSL-terminating HTTP proxy
    # See also: https://twistedmatrix.com/trac/ticket/5807
    hostname, port = twisted_hostname_port(request)
    is_ssl = twisted_is_secure(request)
    request.setHost(hostname, port, is_ssl)

def twisted_hostname_port(request):
    """
    Conveniently get (host, port) tuple of current request,
    either from "Host" header or from the request object itself.
    """
    host_header = request.getHeader(b'Host')
    if host_header:
        if ':' in host_header:
            hostname, port = host_header.split(b':')
        else:
            is_ssl = twisted_is_secure(request)
            hostname, port = host_header, is_ssl and 443 or 80
    else:
        address = request.getHost()
        hostname, port = address.host, address.port

    return hostname, int(port)

def twisted_is_secure(request):
    return request.isSecure() or request.getHeader('X-Forwarded-Proto') == 'https'

def flatten_request_args(args):
    """
    Flatten Twisted request query parameters.
    """
    result = {}
    for key, value in args.iteritems():
        result[key] = ','.join(value)
    return result

def convert_floats(data, integers=None):
    """
    Convert all numeric values in dictionary to float type.
    """
    integers = integers or []
    delete_keys = []
    for key, value in data.iteritems():
        try:
            if isinstance(value, datetime.datetime):
                continue
            if is_number(value):
                if key in integers:
                    data[key] = int(value)
                else:
                    data[key] = float(value)
            if math.isnan(data[key]):
                delete_keys.append(key)
        except:
            pass

    for key in delete_keys:
        del data[key]

    return data

def is_number(s):
    """
    Check string for being a numeric value.
    http://pythoncentral.io/how-to-check-if-a-string-is-a-number-in-python-including-unicode/
    """
    try:
        float(s)
        return True
    except ValueError:
        pass

    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass

    return False

def handleFailure(f, request=None):
    """
    Handle failure in callback chain, log and respond with traceback.

    See also:
    https://twistedmatrix.com/documents/16.0.0/core/howto/defer.html#errbacks
    """
    if f.type is Error:
        if request:
            request.setResponseCode(int(f.value.status))

        if hasattr(f.value, 'with_traceback'):
            f.with_traceback = f.value.with_traceback

        msg = None
        if isinstance(f.value.response, Failure):
            msg = f.value.response.getErrorMessage()
        elif type(f.value.response) in types.StringTypes:
            msg = f.value.response
        request.messages.append({'type': 'error', 'message': msg})

    else:
        if request:
            request.setResponseCode(http.INTERNAL_SERVER_ERROR)
            request.setHeader('Content-Type', 'text/plain; charset=utf-8')
        f.with_traceback = True

    if hasattr(f, 'with_traceback') and f.with_traceback:

        traceback = f.getTraceback()
        log.error(traceback)
        #f.trap(RuntimeError)
        request.write(traceback.encode('utf-8'))


def slugify_datettime(dstring):
    return arrow.get(dstring).to('utc').format('YYYYMMDDTHHmmss')


def parse_timestamp(timestamp):

    if isinstance(timestamp, text_type):

        # HACK: Assume CET (Europe/Berlin) for human readable timestamp w/o timezone offset
        qualified = any([token in timestamp for token in ['Z', '+', ' CET', ' CEST']])
        if not qualified:
            timestamp += ' CET'

        # Parse datetime string
        # Remark: Maybe use pandas.tseries.tools.parse_time_string?
        # TODO: Cache results of call to gettz to improve performance
        berlin = gettz('Europe/Berlin')
        tzinfos = {'CET': berlin, 'CEST': berlin}
        timestamp = parse(timestamp, tzinfos=tzinfos)

    return timestamp



# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from twisted.logger import Logger
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from kotori.bus.wamp import WampSessionMixin, WampBus
from kotori.daq.storage.influx import InfluxDBAdapter

log = Logger()

class InfluxDatabaseService(WampSessionMixin, ApplicationSession):
    """An application component for logging telemetry data to InfluxDB databases"""

    def onJoin(self, details):
        log.info("Realm joined (WAMP session started).")

        # subscribe to telemetry data channel
        self.subscribe(self.receive, u'de.elmyra.kotori.telemetry.data')

        try:
            self.startDatabase()
        except:
            log.failure(u'Problem starting InfluxDB database adapter')

        #self.leave()

    def startDatabase(self):

        log.info('InfluxDB host={host}, version={version}'.format(
            host=self.config.extra['influxdb']['host'],
            version=self.config.extra['influxdb'].get('version', '0.9')))

        self.influx = InfluxDBAdapter(
            settings = self.config.extra['influxdb'],
            database = self.config.extra['hydro2motion']['database'])

    def onLeave(self, details):
        log.info("Realm left (WAMP session ended).")
        ApplicationSession.onLeave(self, details)

    def onDisconnect(self):
        log.info("Transport disconnected.")
        #reactor.stop()


    def receive(self, data):
        #print "RECEIVE:", data

        # decode wire data
        payload = data.split(';')
        #print 'payload:', payload

        try:
            MSG_ID          = int(payload[0])
            V_FC            = float(payload[1]) * 0.001
            V_CAP           = float(payload[2]) * 0.001
            A_ENG           = float(payload[3]) * 0.001
            A_CAP           = float(payload[4]) * 0.001
            T_O2_In         = float(payload[5]) * 0.1
            T_O2_Out        = float(payload[6]) * 0.1
            T_FC_H2O_Out    = float(payload[7]) * 0.1
            Water_In        = float(payload[8])
            Water_Out       = float(payload[9])
            Master_SW       = bool(payload[10])
            CAP_Down_SW     = bool(payload[11])
            Drive_SW        = bool(payload[12])
            FC_state        = bool(payload[13])
            Mosfet_state    = bool(payload[14])
            Safety_state    = bool(payload[15])
            Air_Pump_load   = float(payload[16]) * 0.1
            Mosfet_load     = float(payload[17]) * 0.1
            Water_Pump_load = float(payload[18]) * 0.1
            Fan_load        = float(payload[19]) * 0.1
            Acc_X           = float(payload[20]) * 0.001
            Acc_Y           = float(payload[21]) * 0.001
            Acc_Z           = float(payload[22]) * 0.001
            H2_flow         = float(payload[23]) * 0.001
            GPS_X           = float(payload[24]) * 0.01
            GPS_Y           = float(payload[25]) * 0.01
            GPS_Z           = float(payload[26]) * 0.01
            GPS_Speed       = float(payload[27]) * 0.01
            V_Safety        = float(payload[28]) * 0.001
            H2_Level        = int(payload[29])
            O2_calc         = float(payload[30]) * 0.01
            lat             = float(payload[31])
            lng             = float(payload[32])
            P_In            = A_CAP * V_FC
            P_Out           = A_ENG * V_CAP

            # store data to database
            if self.influx:

                success = self.influx.write_points(
                   'telemetry',
                   ["MSG_ID", "V_FC", "V_CAP", "A_ENG", "A_CAP", "T_O2_In", "T_O2_Out", "T_FC_H2O_Out", "Water_In", "Water_Out", "Master_SW", "CAP_Down_SW", "Drive_SW", "FC_state", "Mosfet_state", "Safety_state", "Air_Pump_load", "Mosfet_load", "Water_Pump_load", "Fan_load", "Acc_X", "Acc_Y", "Acc_Z", "H2_flow", "GPS_X", "GPS_Y", "GPS_Z", "GPS_Speed", "V_Safety", "H2_Level", "O2_calc", "lat", "lng", "P_In", "P_Out"],
                   [MSG_ID, V_FC, V_CAP, A_ENG, A_CAP, T_O2_In, T_O2_Out, T_FC_H2O_Out, Water_In, Water_Out, Master_SW, CAP_Down_SW, Drive_SW, FC_state, Mosfet_state, Safety_state, Air_Pump_load, Mosfet_load, Water_Pump_load, Fan_load, Acc_X, Acc_Y, Acc_Z, H2_flow, GPS_X, GPS_Y, GPS_Z, GPS_Speed, V_Safety, H2_Level, O2_calc, lat, lng, P_In, P_Out]
                )
                log.info("Saved event to InfluxDB: %s" % success)


        except ValueError:
            log.warn('Could not decode data: {}'.format(data))


def h2m_boot_influx_database(settings, debug=False, trace=False):

    websocket_uri = unicode(settings.wamp.uri)

    log.info(u'Starting InfluxDB database service, connecting to WAMP broker "{}"'.format(websocket_uri))

    def application_factory():
        return ApplicationRunner(
            websocket_uri, u'kotori-realm',
            extra={'influxdb': dict(settings.influxdb), 'hydro2motion': dict(settings.hydro2motion)})

    wamp_bus = WampBus(uri=websocket_uri, application_factory=application_factory, session_factory=InfluxDatabaseService)
    wamp_bus.connect()



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@getkotori.org>
from bunch import Bunch
from twisted.application.service import MultiService
from twisted.logger import Logger
from kotori.daq.services import MultiServiceMixin
from kotori.util.twisted import SmartTask

log = Logger()


class DashboardRefreshTamingService(MultiServiceMixin, MultiService):
    """
    Service for taming the refresh interval of Grafana dashboards.
    """

    def __init__(self, channel=None, preset='standard'):
        self.channel = channel or Bunch()
        self.preset = preset
        MultiService.__init__(self)

    def startService(self):
        self.setupService()
        self.log(log.info, u'Starting')

        # Start ourselves as Twisted Service
        MultiService.startService(self)

        # Start all worker tasks
        self.startTasks()

    def setupService(self):
        log.info(u'Setting up {name}', name=self.logname)
        #self.settings = self.parent.settings

    def startTasks(self):
        # Start tamer task
        try:
            self.tamer_start(now=True)
        except Exception as ex:
            log.failure('Starting dashboard refresh interval taming task failed: {ex}\n{log_failure}', ex=ex)

    def tamer_start(self, now=False):
        one_day = 60 * 60 * 24
        self.tamer = SmartTask(worker=self.tamer_process, interval=one_day, onerror='restart')
        self.tamer.start(now=now)

    def tamer_process(self):
        self.parent.tame_refresh_interval(preset=self.preset, force=False)



# -*- coding: utf-8 -*-
# (c) 2015-2016 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
import sys
import socket
import logging
from binascii import unhexlify
from urlparse import urlparse
from tabulate import tabulate
from collections import OrderedDict
from kotori.util.configuration import read_list

logger = logging.getLogger(__name__)

CONFIG_CHANNEL_PREFIX = 'lst-'

def sanitize_channel_label(label):
    return label.replace(CONFIG_CHANNEL_PREFIX, '')

def compute_channel_label(name):
    return CONFIG_CHANNEL_PREFIX + name

def lst_channels(config):
    channel_labels = read_list(config['lst']['channels'])
    channel_infos = []
    for channel_label in channel_labels:
        channel_settings = config[channel_label]
        channel_info = get_channel_info(channel_label, channel_settings)
        channel_infos.append(channel_info)
    print tabulate(channel_infos, headers='keys')

def get_channel_info(channel_label, channel_settings):
    channel = OrderedDict()
    channel['name']         = sanitize_channel_label(channel_label)
    #channel['label']        = channel_label
    channel['udp port']     = channel_settings['udp_port']
    channel['wamp topic']   = channel_settings['wamp_topic']
    channel['header files'] = channel_settings['header_files']
    channel['path']         = os.path.abspath(channel_settings['include_path'])
    return channel

def lst_message(channel, adapter, options):

    target = options.get('--target')
    struct_name = options.get('<name>')
    payload_ascii = options.get('<payload>')

    if options.get('info'):
        print
        if struct_name:
            try:
                struct_adapter = adapter.struct_registry.get(struct_name)
                struct_adapter.print_schema()
            except KeyError:
                logger.error('Struct "{struct_name}" not found in channel "{channel_name}"'.format(
                    struct_name=struct_name, channel_name=channel.name))
                sys.exit(1)
        else:
            channel_info = get_channel_info(channel.label, channel.settings)
            struct_infos = adapter.struct_registry.get_metadata()

            print 'Channel information'; print
            print tabulate(zip(channel_info.keys(), channel_info.values()))
            print; print

            print 'Struct information in "{}"'.format(channel.settings.header_files); print
            print tabulate(struct_infos, headers='keys')

    elif options.get('decode'):
        try:
            struct = decode_payload(adapter, payload_ascii)
            adapter.pprint(struct)
        except KeyError as ex:
            message = 'Decoding binary data "{}" to struct failed.'.format(payload_ascii)
            message += ' ' + ex.message
            logger.error(message)
            sys.exit(1)

    elif options.get('transform'):
        struct = decode_payload(adapter, payload_ascii)
        data = adapter.transform(struct)
        adapter.pprint(struct, data=data)

    elif options.get('send'):
        uri = urlparse(target)
        if uri.scheme == 'udp':

            struct = decode_payload(adapter, payload_ascii)
            #messenger.pprint(struct)

            # send message via UDP
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.sendto(struct._dump_(), (uri.hostname, uri.port))
            logger.info('Message "{}" sent to "{}"'.format(payload_ascii, target))

        else:
            raise ValueError('Can not send message to target "{}", unknown protocol "{}"'.format(target, uri.scheme))


def read_payload(payload):
    # decode data payload from 8-bit clean hex format, e.g. ``0x05022a0021``
    logger.debug('Decoding payload "{}"'.format(payload))
    if payload.startswith('0x'):
        payload = payload.replace('0x', '')
        payload = unhexlify(payload)
    else:
        raise ValueError('Can not decode "{}"'.format(payload))

    return payload


def decode_payload(adapter, payload_ascii):

    # decode from 8-bit clean hex format, e.g. ``0x05022a0021``
    payload = read_payload(payload_ascii)

    # decode binary message
    struct = adapter.decode(payload)

    return struct



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from binascii import hexlify
from cornice.util import to_list
from twisted.logger import Logger
from twisted.internet.protocol import DatagramProtocol
from twisted.internet.defer import inlineCallbacks

class UdpBusForwarder(DatagramProtocol):
    # https://twistedmatrix.com/documents/15.0.0/core/howto/udp.html

    def __init__(self, bus, topic, transform, logger):
        self.bus = bus
        self.topic = topic
        self.transform = to_list(transform)
        self.log = logger or Logger()

        # TODO sanity checks

    @inlineCallbacks
    def datagramReceived(self, data, (host, port)):
        self.log.debug(u"Received via UDP from {host}:{port}: {data}", host=host, port=port, data=hexlify(data))

        # ECHO
        #self.transport.write(data, (host, port))

        # apply transformation steps
        data_out = data
        for transform in self.transform:
            data_out = transform(data_out)

        # forward
        self.log.debug(u"Publishing to topic '{topic}' with realm '{realm}': {data}",
            topic=self.topic, realm=self.bus._realm, data=data_out)

        yield self.bus.publish(self.topic, data_out)



# -*- coding: utf-8 -*-
import types
import pyclibrary
import ctypes
from ctypes import Union, Structure

# taken verbatim from pyclibrary.backends.ctypes.CTypesCLibrary._get_struct
# 2015-11-07 improve _fields_ and _defaults_ building

class StructureWithDefaults(Structure):

    def __init__(self, **kwargs):
        """
        Ctypes.Structure with integrated default values.
        https://stackoverflow.com/questions/7946519/default-values-in-a-ctypes-structure/25892189#25892189

        :param kwargs: values different to defaults
        :type kwargs: dict
        """

        # sanity checks
        defaults = type(self)._defaults_
        assert type(defaults) is types.DictionaryType

        # use defaults, but override with keyword arguments, if any
        values = defaults.copy()
        for (key, val) in kwargs.items():
            values[key] = val

        # appropriately initialize ctypes.Structure
        #super().__init__(**values)                     # Python 3 syntax
        return Structure.__init__(self, **values)       # Python 2 syntax

    # http://stackoverflow.com/questions/1825715/how-to-pack-and-unpack-using-ctypes-structure-str/1827666#1827666
    # https://wiki.python.org/moin/ctypes
    def _dump_(self):
        right = ctypes.sizeof(self) - 1
        return buffer(self)[:right]

    def _load_(self, bytes):
        fit = min(len(bytes), ctypes.sizeof(self))
        ctypes.memmove(ctypes.addressof(self), bytes, fit)


def _get_struct(self, str_type, str_name):
    if str_name not in self._structs_:

        str_name = self._resolve_struct_alias(str_type, str_name)

        # Pull struct definition
        defn = self._defs_[str_type][str_name]

        # create ctypes class
        defs = defn['members'][:]
        if str_type == 'structs':
            class s(StructureWithDefaults):
                def __repr__(self):
                    return "<ctypes struct '%s'>" % str_name
                def _name_(self):
                    return str_name
        elif str_type == 'unions':
            class s(Union):
                def __repr__(self):
                    return "<ctypes union '%s'>" % str_name
                def _name_(self):
                    return str_name

        # Must register struct here to allow recursive definitions.
        self._structs_[str_name] = s

        if defn['pack'] is not None:
            s._pack_ = defn['pack']

        # Assign names to anonymous members
        members = []
        anon = []
        for i, d in enumerate(defs):
            if d[0] is None:
                c = 0
                while True:
                    name = 'anon_member%d' % c
                    if name not in members:
                        d[0] = name
                        anon.append(name)
                        break
            members.append(d[0])

        s._anonymous_ = anon

        # Handle bit field specifications, ctypes only supports bit fields
        # for integer but I am not sure how to test for it in a nice
        # fashion.

        # compute fields
        _fields = []
        for m in defs:
            if len(m) <= 3:
                # the _third_ part of the member definition tuple contains the default value,
                # which should _not_ get used when defining the bare _field_
                t = (m[0], self._get_type(m[1]))

            elif len(m) == 4:
                # the _fourth_ part of the member definition tuple contains the bit width,
                # which _should_ be used when defining the bare _field_
                t = (m[0], self._get_type(m[1]), m[3])

            _fields.append(t)

        # compute defaults
        # the _third_ slot of the member definition tuple contains the default value
        _defaults = {}
        for m in defs:
            if len(m) >= 3:
                key = m[0]
                value = m[2]
                if value:
                    _defaults[key] = value

        s._fields_ = _fields
        s._defaults_ = _defaults

    return self._structs_[str_name]


def monkeypatch_pyclibrary_ctypes_struct():
    pyclibrary.backends.ctypes.CTypesCLibrary._get_struct = _get_struct



# -*- coding: utf-8 -*-
# (c) 2014-2018 Andreas Motl <andreas@getkotori.org>
import re
import os
import sys
import shelve
import socket
import logging
import types

import json_store
from uuid import uuid4
from appdirs import user_data_dir
from datetime import timedelta

from bunch import Bunch

logger = logging.getLogger()


def slm(message):
    """sanitize log message"""
    return unicode(message).replace('{', '{{').replace('}', '}}')


class Singleton(object):
    """
    Singleton class by Duncan Booth.
    Multiple object variables refers to the same object.
    http://www.suttoncourtenay.org.uk/duncan/accu/pythonpatterns.html
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance


class ConfigStoreShelve(dict):

    store = None

    def __init__(self):
        if not ConfigStore.store:
            print "ConfigStoreShelve.__init__"
            self.app_data_dir = user_data_dir('kotori', 'elmyra')
            if not os.path.exists(self.app_data_dir):
                os.makedirs(self.app_data_dir)
            self.config_file = os.path.join(self.app_data_dir, 'config')
            ConfigStore.store = shelve.open(self.config_file, writeback=True)

    def has_key(self, key):
        return ConfigStore.store.has_key(key)

    def __getitem__(self, key):
        print 'ConfigStoreShelve.__getitem__'
        return ConfigStore.store[key]

    def __setitem__(self, key, value):
        print 'ConfigStoreShelve.__setitem__', key, value
        ConfigStore.store[key] = value
        ConfigStore.store.sync()


class ConfigStoreJson(dict):

    store = None

    def __init__(self):
        if not ConfigStoreJson.store:
            #print "ConfigStoreJson.__init__"
            self.app_data_dir = user_data_dir('kotori-daq', 'Elmyra')
            logger.debug("ConfigStoreJson app_data_dir: {}".format(self.app_data_dir))
            if not os.path.exists(self.app_data_dir):
                os.makedirs(self.app_data_dir)
            self.config_file = os.path.join(self.app_data_dir, 'config.json')
            ConfigStoreJson.store = json_store.open(self.config_file)

    def has_key(self, key):
        return ConfigStoreJson.store.has_key(key)

    def __getitem__(self, key):
        #print 'ConfigStoreJson.__getitem__'
        return ConfigStoreJson.store[key]

    def __setitem__(self, key, value):
        #print 'ConfigStoreJson.__setitem__', key, value
        ConfigStoreJson.store[key] = value
        ConfigStoreJson.store.sync()


class NodeId(Singleton):

    config = None
    NODE_ID = 'NODE_UNKNOWN'

    def __init__(self):
        if not self.config:
            self.config = ConfigStoreJson()
        if not self.config.has_key('uuid'):
            self.config['uuid'] = str(uuid4())
        self.NODE_ID = self.config['uuid']
        logger.debug("NODE ID: {}".format(self.NODE_ID))

    def __str__(self):
        return str(self.NODE_ID)


def get_hostname():
    return socket.gethostname()


def setup_logging(level=logging.INFO):
    log_format = '%(asctime)-15s [%(name)-25s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=level)

    # TODO: Control debug logging of HTTP requests through yet another commandline option "--debug-http" or "--debug-requests"
    requests_logger = logging.getLogger('requests')
    requests_logger.setLevel(logging.WARN)


def tdelta(input):
    # https://blog.posativ.org/2011/parsing-human-readable-timedeltas-in-python/
    keys = ["weeks", "days", "hours", "minutes", "seconds"]
    regex = "".join(["((?P<%s>\d+)%s ?)?" % (k, k[0]) for k in keys])
    kwargs = {}
    for k,v in re.match(regex, input).groupdict(default="0").items():
        kwargs[k] = int(v)
    return timedelta(**kwargs)


class SmartBunch(Bunch):

    def dump(self):
        return self.toJSON()

    def pretty(self):
        return self.toJSON(indent=4)

    def prettify(self):
        return self.pretty()

    @classmethod
    def bunchify(cls, x):
        """
        Recursively transforms a dictionary into a SmartBunch via copy.
        Generic "bunchify", also works with descendants of Bunch.
        """
        if isinstance(x, dict):
            return cls( (k, cls.bunchify(v)) for k,v in x.iteritems() )
        elif isinstance(x, (list, tuple)):
            return type(x)( cls.bunchify(v) for v in x )
        else:
            return x


class KeyCache(object):

    def __init__(self):
        self.storage = {}

    # Utility functions for remembering whether the dashboard has been created already.
    # This is important as we would otherwise talk to Grafana for each ingress measurement (on each hit).
    def _get_skip_key(self, *args):
        key_parts = []
        for arg in args:
            if isinstance(arg, types.StringTypes):
                key_parts.append(arg)
            elif isinstance(arg, types.DictionaryType):
                key_parts.append(','.join(arg.keys()))
        skip_key = '-'.join(key_parts)
        return skip_key

    def set(self, *args):
        skip_key = self._get_skip_key(*args)
        self.storage[skip_key] = True

    def exists(self, *args):
        skip_key = self._get_skip_key(*args)
        return skip_key in self.storage



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from pkg_resources import resource_filename
from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource
from pyramid.paster import get_app


def boot_frontend(config, debug=False):
    """
    Boot a Pyramid WSGI application as Twisted component
    """

    http_port = int(config.get('config-web', 'http_port'))
    websocket_uri = unicode(config.get('wamp', 'listen'))

    # https://stackoverflow.com/questions/13122519/serving-pyramid-application-using-twistd/13138610#13138610
    config = resource_filename('kotori.frontend', 'development.ini')
    application = get_app(config, 'main')

    # https://twistedmatrix.com/documents/13.1.0/web/howto/web-in-60/wsgi.html
    resource = WSGIResource(reactor, reactor.getThreadPool(), application)

    reactor.listenTCP(http_port, Site(resource))



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import txmongo
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from twisted.internet.defer import inlineCallbacks
from twisted.internet.interfaces import ILoggingContext
from twisted.python import log
from zope.interface.declarations import implementer

@implementer(ILoggingContext)
class MongoDatabaseService(ApplicationSession):
    """An application component for logging telemetry data to MongoDB databases"""

    #@inlineCallbacks
    #def __init__(self, config):
    #    ApplicationSession.__init__(self, config)

    def logPrefix(self):
        """
        Return a prefix matching the class name, to identify log messages
        related to this protocol instance.
        """
        return self.__class__.__name__

    #@inlineCallbacks
    def onJoin(self, details):
        print("Realm joined (WAMP session started).")

        # subscribe to telemetry data channel
        self.subscribe(self.receive, u'de.elmyra.kotori.telemetry.data')

        self.startDatabase()

        #self.leave()

    #@inlineCallbacks
    def startDatabase(self):
        #self.mongo = yield txmongo.MongoConnection(host='127.0.0.0', port=27017)
        self.mongo = yield txmongo.MongoConnection()

    def onLeave(self, details):
        print("Realm left (WAMP session ended).")
        ApplicationSession.onLeave(self, details)

    def onDisconnect(self):
        print("Transport disconnected.")
        #reactor.stop()

    #@inlineCallbacks
    def receive(self, data):
        #print "RECEIVE:", data

        # decode wire data
        payload = data.split(';')
        try:
            MSG_ID          = int(payload[0])
            V_FC            = int(payload[1])
            V_CAP           = int(payload[2])
            A_ENG           = int(payload[3])
            A_CAP           = int(payload[4])
            T_O2_In         = int(payload[5])
            T_O2_Out        = int(payload[6])
            T_FC_H2O_Out    = int(payload[7])
            Water_In        = int(payload[8])
            Water_Out       = int(payload[9])
            Master_SW       = bool(payload[10])
            CAP_Down_SW     = bool(payload[11])
            Drive_SW        = bool(payload[12])
            FC_state        = bool(payload[13])
            Mosfet_state    = bool(payload[14])
            Safety_state    = bool(payload[15])
            Air_Pump_load   = float(payload[16])
            Mosfet_load     = int(payload[17])
            Water_Pump_load = int(payload[18])
            Fan_load        = int(payload[19])
            Acc_X           = int(payload[20])
            Acc_Y           = int(payload[21])
            Acc_Z           = int(payload[22])
            AUX             = float(payload[23])
            GPS_X           = int(payload[24])
            GPS_Y           = int(payload[25])
            GPS_Z           = int(payload[26])
            GPS_Speed       = int(payload[27])
            V_Safety        = int(payload[28])
            H2_Level        = int(payload[29])
            O2_calc         = float(payload[30])
            lat             = float(payload[31])
            lng             = float(payload[32])

            # store data to database
            if self.mongo:
                telemetry = self.mongo.kotori.telemetry
                yield telemetry.insert(dict(MSG_ID = MSG_ID, V_FC = V_FC, V_CAP = V_CAP, A_ENG = A_ENG, A_CAP = A_CAP, T_O2_In = T_O2_In, T_O2_Out = T_O2_Out, T_FC_H2O_Out = T_FC_H2O_Out, Water_In = Water_In, Water_Out = Water_Out, Master_SW = Master_SW, CAP_Down_SW = CAP_Down_SW, Drive_SW = Drive_SW, FC_state = FC_state, Mosfet_state = Mosfet_state, Safety_state = Safety_state, Air_Pump_load = Air_Pump_load, Mosfet_load = Mosfet_load, Water_Pump_load = Water_Pump_load, Fan_load = Fan_load, Acc_X = Acc_X, Acc_Y = Acc_Y, Acc_Z = Acc_Z, AUX = AUX, GPS_X = GPS_X, GPS_Y = GPS_Y, GPS_Z = GPS_Z, GPS_Speed = GPS_Speed, V_Safety = V_Safety, H2_Level = H2_Level, O2_calc = O2_calc, lat = lat, lng = lng))

        except ValueError:
            print('Could not decode data: {}'.format(data))


def boot_mongo_database(websocket_uri, debug=False, trace=False):

    print 'INFO: Starting mongo database service, connecting to broker', websocket_uri

    runner = ApplicationRunner(websocket_uri, u'kotori-realm', debug=trace, debug_wamp=debug, debug_app=debug)
    runner.run(MongoDatabaseService, start_reactor=False)



# -*- coding: utf-8 -*-
# (c) 2014-2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
#
# derived from https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/pubsub/simple/example2/server.py
import sys
import datetime
from copy import deepcopy
from urlparse import urlparse, parse_qs
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from autobahn.twisted.websocket import WampWebSocketClientProtocol, WampWebSocketClientFactory
from autobahn.twisted.websocket import WampWebSocketServerProtocol, WampWebSocketServerFactory
from twisted.python import log
from twisted.internet import reactor
from kotori.util.common import ConfigStoreJson


client = None


class NodeRegistry(object):

    def __init__(self):
        self.nodes = {}
        self.sessions = {}
        self.config = ConfigStoreJson()
        #print '----------------------', self.config.store
        #print '----------------------', self.config.get('nodes')
        #self.nodes = self.config.get('nodes', {})
        #print '======================', self.nodes

    def persist(self):
        print "============= persist", self.nodes
        #self.config['nodes'].update(self.nodes)
        #del self.config['nodes']
        self.config['nodes'] = {'haha': {'hehe': 'dscsdcsdcsdcc'}}
        #self.config['nodes'] = self.nodes.copy()
        self.config['nodes2'] = deepcopy(self.nodes)
        self.config['ttt'] = 'huhu'

        #   self.config['nodes']['abc'] = 'def2'
        #self.nodes.setdefault('ghi', {})['kli'] = 'DEFUNCT'
        #self.config['ghi']['jkl'] = 'defunct'

    def register(self, node_id, hostname=None):
        print "NodeRegistry.register:", node_id
        self.nodes[node_id] = {'hostname': hostname}
        self.persist()
        client.publish('http://kotori.elmyra.de/dashboard#update', None)

    def unregister(self, node_id):
        print "NodeRegistry.unregister:", node_id
        try:
            del self.nodes[node_id]
            self.persist()
        except KeyError:
            pass
        client.publish('http://kotori.elmyra.de/dashboard#update', None)

    #@exportRpc
    def get_nodes(self):
        print "NodeRegistry.get_nodes:", self.nodes
        return self.nodes

    #@exportRpc
    def set_node_label(self, node_id, label):
        print "NodeRegistry.set_node_label:", node_id, label
        self.nodes.setdefault(node_id, {})['label'] = label
        print self.nodes[node_id]
        self.persist()

#registry = NodeRegistry()


class MasterServerProtocol(WampWebSocketServerProtocol):

    def onSessionOpen(self):
        self.registerForRpc(registry, "http://kotori.elmyra.de/registry#")
        self.registerForPubSub("http://kotori.elmyra.de/broadcast#", True)
        self.registerForPubSub("http://kotori.elmyra.de/dashboard#", True)
        self.registerForPubSub("http://kotori.elmyra.de/presence", True)
        self.registerForPubSub("http://kotori.elmyra.de/node/", True)



class MasterServerFactory(WampWebSocketServerFactory):

    protocol = MasterServerProtocol

    def __init__(self, *args, **kwargs):
        WampServerFactory.__init__(self, *args, **kwargs)
        self.setProtocolOptions(allowHixie76 = True)

    def onClientSubscribed(self, proto, topic):
        print "subscribed:  ", proto, topic
        uri = urlparse(topic)
        if uri.path == '/presence':
            params = parse_qs(uri.query)
            node_id = params['node_id'][0]
            hostname = params['hostname'][0]
            registry.register(node_id, hostname)

    def onClientUnsubscribed(self, proto, topic):
        print "unsubscribed:", proto, topic
        uri = urlparse(topic)
        if uri.path == '/presence':
            params = parse_qs(uri.query)
            node_id = params['node_id'][0]
            registry.unregister(node_id)


class MasterClientProtocol(WampWebSocketClientProtocol):
    def onSessionOpen(self):
        global client
        client = self

class MasterClientFactory(WampWebSocketClientFactory):
    protocol = MasterClientProtocol


class Component(ApplicationSession):

    """
    A simple time service application component.
    """

    def onJoin(self, details):

        def utcnow():
            now = datetime.datetime.utcnow()
            return now.strftime("%Y-%m-%dT%H:%M:%SZ")

        self.register(utcnow, 'com.timeservice.now')


def boot_master(websocket_uri, debug=False, trace=False):

    print 'INFO: Starting WebSocket master service on', websocket_uri
    """
    factory = MasterServerFactory(websocket_uri, debugWamp = True)
    websocket.listenWS(factory)

    client_factory = MasterClientFactory(websocket_uri, debug=False, debugCodePaths=False, debugWamp=debug, debugApp=False)
    websocket.connectWS(client_factory)
    """

    runner = ApplicationRunner(websocket_uri, u'kotori-realm', debug=trace, debug_wamp=debug, debug_app=debug)
    runner.run(Component, start_reactor=False)


def run():

    if len(sys.argv) > 1 and sys.argv[1] == 'debug':
        log.startLogging(sys.stdout)
        debug = True
    else:
        debug = False

    websocket_uri = 'ws://localhost:9000'
    http_port = 35000

    boot_master(websocket_uri, http_port, debug)

    reactor.run()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@getkotori.org>
from __future__ import absolute_import
from twisted.logger import Logger
from twisted.internet.task import LoopingCall

log = Logger()


class SmartTask(object):

    def __init__(self, worker=None, interval=None, onerror='stop'):

        assert worker, 'The "worker" parameter is missing or None'
        assert interval, 'The "interval" parameter is missing or None'
        assert onerror in ['stop', 'restart'], 'Unknown value "{}" for parameter "onerror"'.format(onerror)

        self.worker = worker
        self.interval = interval
        self.onerror = onerror
        self.task = None

    def start(self, now=False):
        self.task = LoopingCall(self.worker)
        self.deferred = self.task.start(self.interval, now=now)
        self.deferred.addErrback(self.handle_error)

    def handle_error(self, failure, *args):
        log.failure('Task failed:\n{log_failure}', failure=failure)

        # Restart task after error
        if self.onerror == 'restart':
            self.start(now=False)



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl, <andreas@getkotori.org>
from twisted.logger import Logger
from grafana_api_client import GrafanaClient, GrafanaServerError, GrafanaPreconditionFailedError, GrafanaClientError

log = Logger()


class GrafanaApi(object):
    """
    A small wrapper around ``grafana_api_client``.
    https://pypi.python.org/pypi/grafana_api_client
    """

    def __init__(self, host='localhost', port=3000, username='admin', password='admin'):
        self.host = host
        self.port = port
        self.username = username
        self.password = password

        # A GrafanaClient instance
        self.grafana_client = None

        # The uid of the "Instant Dashboards" folder
        self.instant_folder_uid = u'instagraf'
        self.instant_folder_title = u'# Instant Dashboards'

        self.connect()

    def connect(self):
        self.grafana_client = GrafanaClient((self.username, self.password), host=self.host, port=self.port)
        try:
            self.ensure_instant_folder()
        except Exception as ex:
            log.warn(u'Problem creating instant folder: {ex}', ex=ex)

    def ensure_instant_folder(self):
        return self.ensure_folder(uid=self.instant_folder_uid, title=self.instant_folder_title)

    def ensure_folder(self, uid=None, title=None):

        try:
            return self.get_folder(uid=uid)

        except GrafanaClientError as ex:

            if '404' in ex.message or 'not-found' in ex.message or 'Folder not found' in ex.message:
                log.debug(u'Folder with uid="{uid}" not found, creating', uid=uid)
                try:
                    return self.create_folder(uid=uid, title=title)

                except Exception as ex:
                    log.warn(u'Problem creating folder uid="{uid}", title="{title}": {ex}', uid=uid, title=title, ex=ex)

            else:
                log.warn(u'Problem getting folder uid="{uid}"": {ex}', uid=uid, ex=ex)
                raise


    def create_folder(self, uid=None, title=None):
        log.info(u'Creating folder with uid="{uid}" and title="{title}"', uid=uid, title=title)
        data = {}
        if uid: data['uid'] = uid
        if title: data['title'] = title
        try:
            return self.grafana_client.folders.create(**data)

        except GrafanaPreconditionFailedError as ex:
            # Ignore modifications from other users while doing our own
            # TODO: Add some locking mechanisms to protect against these issues
            if 'version-mismatch' in ex.message or 'The folder has been changed by someone else' in ex.message:
                #log.warn('{message}', message=ex.message)
                pass
            else:
                raise

    def get_folder(self, uid):
        log.info('Get folder with uid="{}"'.format(uid))
        data = self.grafana_client.folders[uid].get()
        return data

    def create_datasource(self, name=None, data=None):
        data = data or {}
        data.setdefault('name', name)
        data.setdefault('access', 'proxy')

        name = data['name']

        # TODO: can delete by datasource "id" only. have to inquire using "GET /api/datasources" first
        """
        try:
            logger.info('deleting datasource: {}'.format(name))
            response = self.grafana_client.datasources[name].delete()
            print response
        except GrafanaClientError as ex:
            if '404' in ex.message or 'Dashboard not found' in ex.message:
                logger.warn(slm(ex.message))
            else:
                raise
        """

        try:
            log.info(u'Checking/Creating datasource "{}"'.format(name))
            response = self.grafana_client.datasources.create(**data)
            log.info('response: {response}', response=response)
        except GrafanaServerError as ex:
            if 'Failed to add datasource' in ex.message:
                pass
            else:
                raise
        except GrafanaClientError as ex:
            if 'Data source with same name already exists' in ex.message:
                pass
            else:
                raise

        #print grafana.datasources()


    def create_dashboard(self, dashboard, name=None, delete=False):

        if not name:
            name = dashboard.get_title()

        name = self.format_dashboard_name(name)

        if delete:
            try:
                log.info(u'Deleting dashboard "{}"'.format(name))
                response = self.grafana_client.dashboards.db[name].delete()
                log.info(u'Grafana response: {response}', response=response)

            except GrafanaClientError as ex:
                if '404' in ex.message or 'Dashboard not found' in ex.message:
                    log.warn(u'{message}', message=ex.message)
                else:
                    raise

        try:
            log.info(u'Creating/updating dashboard "{}"'.format(name))
            dashboard_payload = dashboard.wrap_api()
            response = self.grafana_client.dashboards.db.create(**dashboard_payload)
            log.info(u'Grafana response: {response}', response=response)

        except GrafanaPreconditionFailedError as ex:
            if 'name-exists' in ex.message or 'A dashboard with the same name already exists' in ex.message:
                log.warn(u'{message}', message=ex.message)
            else:
                raise

        try:
            log.info(u'Checking dashboard "{}"'.format(name))
            dashboard = self.grafana_client.dashboards.db[name].get()
        except GrafanaClientError as ex:
            if '404' in ex.message or 'Dashboard not found' in ex.message:
                log.warn(u'{message}', message=ex.message)
            else:
                raise

    def get_dashboard(self, name):
        try:
            name = self.format_dashboard_name(name)
            log.info(u'Getting dashboard "{}"'.format(name))
            dashboard = self.grafana_client.dashboards.db[name].get()
            return dashboard['dashboard']
        except GrafanaClientError as ex:
            if '404' in ex.message or 'Dashboard not found' in ex.message:
                log.info(u'{message}', message=ex.message)
            else:
                raise

    def get_dashboard_by_uid(self, uid):
        """
        Will return the dashboard given the dashboard unique identifier (uid).
        http://docs.grafana.org/http_api/dashboard/#get-dashboard-by-uid

        :param uid: Unique dashboard identifier (uid)
        :return:    Dashboard data structure as dictionary
        """
        return self.grafana_client.dashboards.uid[uid].get()

    def format_dashboard_name(self, name):
        return name.replace(' ', '-')

    def demo(self):
        print grafana.org()
        #    {"id":1,"name":"Main Org."}
        #client.org.replace(name="Your Org Ltd.")
        #    {"id":1,"name":"Your Org Ltd."}

    def get_dashboards(self):
        return self.grafana_client.search(type='dash-db')



# -*- coding: utf-8 -*-
# (c) 2014 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import os
import sys
import urllib

def tts_say(message, language='de'):

    # FIXME: add queue here
    print "say:", message

    # Google Translate TTS
    # FIXME: urlescape "message"
    tts_url = u'http://translate.google.com/translate_tts?tl={language}&q={message}'.format(language=language, message=urllib.quote(message.encode('utf8')))
    more_args = u''
    if sys.platform.startswith('linux'):
        # TODO: dynamic configuration of output device
        more_args += u'-ao alsa:device=hw=1.0'

    user_agent = u"-http-header-fields 'User-Agent: Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.63 Safari/537.36'"
    command = u"mplayer -really-quiet -noconsolecontrols {user_agent} {more_args} '{tts_url}'".format(**locals())
    #command = command.encode('utf8')
    print command
    # FIXME: don't do this synchronously
    # FIXME: show errors (stdout/stderr) if command fails
    os.system(command)



# -*- coding: utf-8 -*-
# (c) 2014 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from twisted.internet import reactor
try:
    import Adafruit_BBIO.GPIO as GPIO
except Exception as ex:
    print "ERROR: No Adafruit_BBIO module installed:", ex

class GpioInput(object):

    def __init__(self, port, callback):
        self.port = port
        self.callback = callback

        GPIO.setup(self.port, GPIO.IN)

        print "DETECT EVENTS:", self.port

        # GPIO.RISING, GPIO.FALLING, GPIO.BOTH
        GPIO.remove_event_detect(self.port)
        GPIO.add_event_detect(self.port, GPIO.RISING, callback=self.check_rising, bouncetime=300)
        #GPIO.add_event_detect(self.port, GPIO.BOTH, callback=self.check_rising)
        #GPIO.add_event_detect(self.port, GPIO.FALLING, callback=self.check_falling)
        #GPIO.add_event_detect(self.port, GPIO.FALLING, self.check_falling)

    def check_rising(self, port):
        #print "check_rising:", port, self.port
        if port == self.port:
            reactor.callFromThread(self.callback, port)

    # doesn't work with Adafruit GPIO
    """
    def check_falling(self, port):
        if port == self.port:
            #self.callback(port)
            print "================= FALLING"
    """



class GpioOutput(object):

    def __init__(self, port):
        self.port = port
        GPIO.setup(self.port, GPIO.OUT)

    def on(self):
        self.high()

    def off(self):
        self.low()

    def low(self):
        GPIO.output(self.port, GPIO.LOW)

    def high(self):
        GPIO.output(self.port, GPIO.HIGH)



# -*- coding: utf-8 -*-
# (c) 2014-2017 Andreas Motl, <andreas@getkotori.org>
import os
import logging
from glob import glob
from bunch import Bunch
from cornice.util import to_list
from ConfigParser import ConfigParser

log = logging.getLogger()

def get_configuration_file(configfile=None):
    # compute configuration file
    if not configfile:
        configfile = os.environ.get('KOTORI_CONFIG')

    if not configfile:
        raise ValueError('No configuration file, either use --config=/path/to/kotori.ini or set KOTORI_CONFIG environment variable')

    log.info('Root configuration file is {}'.format(configfile))
    return configfile

def get_configuration(*args):
    config_files = []
    # TODO: Only use when logged in interactively?
    config_files += [os.path.expanduser('~/.kotori.ini')]
    config_files += list(args)
    log.info('Requested configuration files: {}'.format(make_list(config_files)))
    config, used = read_config(config_files, kind=Bunch)
    if config:
        if 'main' in config and 'include' in config.main:
            includes = read_list(config.main.include)
            for include in includes:
                if '*' in include or '?' in include:
                    config_files += glob(include)
                else:
                    config_files.append(include)
            log.info('Expanded configuration files:  {}'.format(make_list(config_files)))
            config, used = read_config(config_files, kind=Bunch)
        log.info('Used configuration files:      {}'.format(make_list(used)))
        return config
    else:
        msg = u'Could not read settings from configuration files: {}'.format(config_files)
        log.critical(msg)
        raise ValueError(msg)

def read_config(configfiles, kind=None):
    configfiles_requested = to_list(configfiles)
    config = ConfigParser()
    configfiles_used = config.read(configfiles_requested)
    settings = convert_config(config, kind=kind)
    return settings, configfiles_used

def convert_config(config, kind=None):
    """
    Serialize section-based ConfigParser contents
    into nested dict or other dict-like thing.
    """
    kind = kind or dict
    if isinstance(config, ConfigParser):
        config_dict = kind()
        for section in config.sections():
            config_dict[section] = kind(config.items(section))
        return config_dict
    else:
        return config

def read_list(string, separator=u',', empty_elements=True):
    data = map(unicode.strip, string.split(separator))
    if empty_elements == False:
        data = filter(lambda x: bool(x), data)
    return data

def make_list(items, separator=u', '):
    return separator.join(items)


def apply_default_settings(settings):

    # MQTT setting defaults
    settings.mqtt.setdefault('host', u'localhost')
    settings.mqtt.setdefault('port', u'1883')
    settings.mqtt.setdefault('debug', u'false')




# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl <andreas.motl@elmyra.de>
import types
import tempfile
from pprint import pprint
from string import Template
from pkg_resources import resource_string
from twisted.web.template import renderElement
from twisted.logger import Logger
from kotori.io.export.html import DygraphsPage
from kotori.io.protocol.util import get_data_uri
from kotori.io.export.util import dataframe_index_to_column, dataframe_wide_to_long_indexed, dataframe_index_and_sort
from kotori.io.export.util import matplotlib_locator_formatter

log = Logger()

class UniversalPlotter(object):
    """
    Universal plotter for timeseries data.

    Render pandas DataFrame to different timeseries plots.
    See also: http://pandas.pydata.org/pandas-docs/stable/cookbook.html#cookbook-plotting

    Tabular data:

        - CSV
        - JSON
        - HTML
        - Excel (XLSX)
        - HDF5
        - NetCDF
        - DataTables HTML widget

    Timeseries plots:

        - [PNG]  matplotlib
        - [PNG]  ggplot
        - [HTML] dygraphs
        - [HTML] Bokeh
        - [HTML] Vega/Vincent

    """

    def __init__(self, bucket, dataframe):
        self.bucket = bucket
        self.request = bucket.request
        self.dataframe = dataframe

    def render(self, format, kind=None, buffer=None):
        if format == 'png':
            return self.render_png(buffer)
        elif format == 'html':
            return self.render_html(kind)
        elif format == 'json':
            return self.render_json(kind)

    def render_png(self, buffer):
        """
        Render timeseries plots as PNG images.
        """

        bucket = self.bucket

        import matplotlib.font_manager
        matplotlib.font_manager.findSystemFonts(fontpaths=None, fontext='ttf')

        import matplotlib
        try:
            matplotlib.use('agg')
        except:
            pass

        import matplotlib.pyplot as plt

        df = self.dataframe
        #df = df.set_index(['time'])

        # Compute datetime range boundaries
        datetime_min = min(df.time)
        datetime_max = max(df.time)
        datetime_delta = datetime_max - datetime_min
        #xmin = pd.to_datetime('2016-05-01')
        #xmax = pd.to_datetime('2016-08-01')

        renderer = bucket.tdata.get('renderer', 'matplotlib')
        if renderer == 'matplotlib':

            # Bring DataFrame into appropriate format
            df = dataframe_index_and_sort(df, 'time')

            # Propagate non-null values forward or backward, otherwise
            # matplotlib would not plot the sparse data frame properly.
            # With time series data, using pad/ffill is extremely common so that the “last known value” is available at every time point.
            # http://pandas.pydata.org/pandas-docs/stable/missing_data.html#filling-missing-values-fillna
            df.fillna(method='pad', inplace=True)

            # Make plots of DataFrame using matplotlib / pylab.
            # http://matplotlib.org/
            # http://pandas.pydata.org/pandas-docs/version/0.13.1/visualization.html
            # http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.plot.html
            # https://markthegraph.blogspot.de/2015/05/plotting-time-series-dataframes-in.html

            if 'style' in bucket.tdata and bucket.tdata.style:
                try:
                    plt.style.use(bucket.tdata.style)
                except Exception:
                    error_message = u'# Unknown style "{style_name}", available styles: {available}'.format(
                        style_name=bucket.tdata.style, available=plt.style.available)
                    log.error(error_message)
                    return self.request.error_response(bucket, error_message)


            # Basic plotting
            #df.plot()
            #plt.savefig(buffer)


            # Advanced plotting
            ax = df.plot()
            fig = ax.get_figure()

            # Figure heading
            title = fig.suptitle(bucket.title.human, fontsize=12)
            #fig.tight_layout(pad=1.5)

            # Axis and tick labels
            ax.set_xlabel('Time')
            ax.set_ylabel('Value')
            ax.tick_params(axis='x', labelsize='smaller')

            # Grid and legend
            # http://matplotlib.org/users/legend_guide.html
            # http://matplotlib.org/examples/pylab_examples/legend_demo3.html
            ax.grid(True)

            legend_params = dict(ncol=1, loc='center left', bbox_to_anchor=(1, 0.5), fontsize='small', shadow=True, fancybox=True)
            legend = ax.legend(**legend_params) # title='Origin'
            #ax.legend(**legend_params) # title='Origin'

            # Sort list of legend labels
            # http://stackoverflow.com/questions/22263807/how-is-order-of-items-in-matplotlib-legend-determined/27512450#27512450


            # Axis formatting
            #ax.xaxis_date()
            #ax.autoscale_view()


            # Compute appropriate locator and formatter
            locator, formatter = matplotlib_locator_formatter(datetime_delta, span=1)

            #ax.xaxis.set_major_locator(locator)
            ax.xaxis.set_major_formatter(formatter)

            # Figure formatting
            fig.autofmt_xdate()

            # http://stackoverflow.com/questions/10101700/moving-matplotlib-legend-outside-of-the-axis-makes-it-cutoff-by-the-figure-box/10154763#10154763
            fig.savefig(buffer, bbox_extra_artists=(title, legend), bbox_inches='tight')

            # TODO: Add annotations
            """
            # https://stackoverflow.com/questions/11067368/annotate-time-series-plot-in-matplotlib
            # https://stackoverflow.com/questions/17891493/annotating-points-from-a-pandas-dataframe-in-matplotlib-plot
            import matplotlib.dates as mdates
            fig = plot.draw()
            ax = fig.axes[0]
            ax.annotate('Test', (mdates.date2num(x[1]), y[1]), xytext=(15, 15),
                textcoords='offset points', arrowprops=dict(arrowstyle='-|>'))

            """


        elif renderer == 'ggplot':

            # https://yhat.github.io/ggplot/notebook.html?page=build/docs/examples/Multiple%20Line%20Plot.html
            # https://stackoverflow.com/questions/23541497/is-there-a-way-to-plot-a-pandas-series-in-ggplot
            # https://stackoverflow.com/questions/24478925/is-it-possible-to-plot-multiline-chart-on-python-ggplot/24479513#24479513

            # https://github.com/yhat/ggplot/blob/master/docs/how-to/Building%20Faceted%20(or%20Trellised)%20Plots.ipynb
            # https://github.com/yhat/ggplot/blob/master/docs/how-to/Annotating%20Plots%20-%20Titles%20and%20Labels.ipynb
            # https://github.com/yhat/ggplot/blob/master/docs/how-to/How%20to%20make%20xkcd%20style%20graphs.ipynb

            from ggplot import ggplot, aes, qplot, geom_line, geom_text, ggtitle, stat_smooth, scale_x_date, date_format, date_breaks
            from ggplot import theme_538, theme_bw, theme_gray, theme_xkcd

            # https://stackoverflow.com/questions/24478925/is-it-possible-to-plot-multiline-chart-on-python-ggplot/24479513#24479513
            # https://stackoverflow.com/questions/23541497/is-there-a-way-to-plot-a-pandas-series-in-ggplot

            # Convert DataFrame from wide to long format, retaining "time" as visible column
            df = dataframe_wide_to_long_indexed(df, 'time')
            dataframe_index_to_column(df, 'time')

            # Compute appropriate locator and formatter
            locator, formatter = matplotlib_locator_formatter(datetime_delta, span=2)

            plot = ggplot(df, aes(x='time', y='value', color='variable'))\
                   + geom_line()\
                   + scale_x_date(limits=(datetime_min, datetime_max), breaks=locator, labels=formatter)\
                   + ggtitle(bucket.title.human)

            # Axis labels
            plot.xlab = 'Time'
            plot.ylab = 'Value'

            # Labs
            #+ stat_smooth(colour='blue', span=0.2) \
            #+ geom_text(aes(x='x', y='y'), label='hello world')
            #+ scale_x_date(limits=(xmin, xmax), breaks=date_breaks('1 hour'), labels=date_format('%Y-%m-%d\n%H:%M'))

            theme_name = bucket.tdata.get('theme')
            # TODO: Switching themes will leak some matplotlib/pyplot properties, postpone to future versions
            if theme_name:
                if type(theme_name) is types.FloatType:
                    theme_name = str(int(theme_name))
                try:
                    theme = eval('theme_' + theme_name)
                    plot += theme()
                except Exception:
                    error_message = u'# Unknown theme "{theme_name}"'.format(theme_name=theme_name)
                    log.error(error_message)
                    return self.request.error_response(bucket, error_message)

            plot.save(buffer)

            # Attempt to reset global matplotlib parameters to get rid of xkcd theme style
            """
            import matplotlib as mpl
            #mpl.rcParams = mpl.rc_params()
            #del mpl.rcParams['path.sketch']
            #del mpl.rcParams['path.effects']
            #mpl.rcParams = mpl.defaultParams.copy()
            #mpl.rcParams.clear()
            #mpl.rcdefaults()
            #mpl.rcParams = mpl.rcParamsOrig
            if 'axes.prop_cycle' in mpl.rcParams:
                del mpl.rcParams['axes.prop_cycle']
            mpl.rcParams.update({'path.sketch': None, 'path.effects': []})
            mpl.rcParams.update(mpl.rc_params())
            """

        elif renderer == 'seaborn':

            # TODO: We don't do statistical plotting yet.

            # https://stanford.edu/~mwaskom/software/seaborn/examples/timeseries_from_dataframe.html
            # https://stanford.edu/~mwaskom/software/seaborn/generated/seaborn.tsplot.html
            import seaborn as sns
            sns.set(style="darkgrid")
            #sns.tsplot(data=gammas, time="timepoint", unit="subject", condition="ROI", value="BOLD signal")
            #print dir(df)
            #df['time'] = pandas.to_datetime(df['time'])
            #df = df.set_index(df.time)
            pprint(df)
            sns.tsplot(data=df, time="time")
            #sns.tsplot(data=df)
            plt.savefig(buffer)

        else:
            error_message = u'# Unknown renderer "{renderer_name}"'.format(renderer_name=renderer)
            log.error(error_message)
            return self.request.error_response(bucket, error_message)

    def render_html(self, kind):
        """
        Render HTML-based timeseries plots for dygraphs, Bokeh and Vega.
        """

        # Variable aliases
        bucket = self.bucket
        df = self.dataframe

        if kind == 'dygraphs':
            # http://dygraphs.com/

            # Compute data_uri, forward "from" and "to" parameters
            data_uri = get_data_uri(bucket, 'data.csv', {'pad': 'true'})

            # Render HTML snippet containing dygraphs widget
            page = DygraphsPage(data_uri=data_uri, bucket=bucket)
            bucket.request.setHeader('Content-Type', 'text/html; charset=utf-8')
            return renderElement(bucket.request, page)

        elif kind == 'bokeh':
            # http://bokeh.pydata.org/

            from bokeh.io import save
            from bokeh.charts import TimeSeries, vplot

            # Propagate non-null values forward or backward, otherwise
            # Bokeh would not plot the sparse data frame properly.
            # With time series data, using pad/ffill is extremely common so that the “last known value” is available at every time point.
            # http://pandas.pydata.org/pandas-docs/stable/missing_data.html#filling-missing-values-fillna
            df.fillna(method='pad', inplace=True)

            # Plot using matplotlib
            # http://bokeh.pydata.org/en/latest/docs/user_guide/compat.html#userguide-compat
            # https://github.com/bokeh/bokeh/tree/master/examples/compat/
            # https://github.com/bokeh/bokeh/blob/master/examples/compat/pandas_dataframe.py
            # https://github.com/bokeh/bokeh/blob/master/examples/compat/ggplot_line.py
            #df.plot()
            #what = mpl.to_bokeh()

            # Plot using Bokeh TimeSeries
            # http://bokeh.pydata.org/en/latest/docs/reference/charts.html#timeseries
            # http://bokeh.pydata.org/en/0.11.1/docs/user_guide/styling.html#location
            linegraph = TimeSeries(df, x='time', title=bucket.title.human, legend="top_left", width=800)

            # Plot TimeSeries object
            what = vplot(linegraph)

            # Render using Bokeh
            t = tempfile.NamedTemporaryFile(suffix='.html', delete=False)
            save(what, filename=t.name, title=bucket.title.human)

            # Forward html to http response
            bucket.request.setHeader('Content-Type', 'text/html; charset=utf-8')
            return t.read()

        elif kind == 'vega':
            # https://github.com/wrobstory/vincent

            # Compute data_uri, forward "from" and "to" parameters
            data_uri = get_data_uri(bucket, 'data.vega.json', {'pad': 'true', 'backfill': 'true'})

            template = Template(str(resource_string('kotori.io.export', 'vega_template.html')))
            bucket.request.setHeader('Content-Type', 'text/html; charset=utf-8')
            response = template.substitute(path=data_uri, title=bucket.title.human)
            return response.encode('utf-8')

    def render_json(self, kind):
        """
        Render JSON chart description for HTML-based timeseries plot Vega, using Vincent.
        """

        # Variable aliases
        bucket = self.bucket
        df = self.dataframe

        if kind == 'vega':
            # https://github.com/wrobstory/vincent
            # https://github.com/wrobstory/vincent/blob/master/examples/line_chart_examples.py
            # https://stackoverflow.com/questions/29288914/how-to-get-vincent-to-display-a-pandas-date-time-axis-correctly
            # https://wrobstory.github.io/2013/04/pandas-vincent-timeseries.html

            from vincent import Line

            # Mungle DataFrame into appropriate format
            df = dataframe_index_and_sort(df, 'time')

            vis = Line(df)
            vis.axis_titles(x='Time', y='Value')
            vis.legend(title='Origin')

            bucket.request.setHeader('Content-Type', 'application/json')
            return vis.to_json()




# -*- coding: utf-8 -*-
# (c) 2016 Andreas Motl, <andreas@getkotori.org>
from bunch import Bunch
from twisted.logger import Logger
from kotori import KotoriBootloader
from kotori.daq.services import RootService
from kotori.util.configuration import read_list

log = Logger()

class CompositeApplication(RootService):
    """
    Application service container root object
    """

    def __init__(self, name=None, application_settings=None, global_settings=None):
        RootService.__init__(self, settings=global_settings)

        # Compute name for Twisted service
        self.name = u'app-{name}'.format(name=name)

        # Make channel object from application settings configuration object
        self.channel = Bunch(**application_settings)

        # Main application components
        for service_reference in read_list(application_settings.services):

            # Load service component
            try:
                service_factory = KotoriBootloader.load_entrypoint(service_reference)
            except Exception as ex:
                log.failure('Error loading composite service component "{service_name}" into "{app_name}":\n{log_failure}"',
                            service_name=service_reference, app_name=name)
                continue

            # Load data processing strategy and graphing components
            # TODO: Review whether this should be per-service or not
            # TODO: Introduce strict/non-strict handling
            strategy_factory = KotoriBootloader.load_entrypoint(application_settings.strategy, onerror='ignore')
            graphing_factory = KotoriBootloader.load_entrypoint(application_settings.graphing, onerror='ignore')

            # Create application service object composed of subsystem components
            # TODO: Bundle all arguments into yet another wrapper object for an universal object factory

            service = service_factory(
                channel  = self.channel,
                strategy = strategy_factory(settings=global_settings),
                graphing = graphing_factory(settings=global_settings, channel=self.channel)
                )

            # Register service component with its container
            self.registerService(service)


def boot(name=None, **kwargs):
    # Bootstrap application composed of components defined by configuration settings
    app = CompositeApplication(name=name, **kwargs)
    return app



# -*- coding: utf-8 -*-
# (c) 2014-2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
#
# derived from https://github.com/tavendo/AutobahnPython/blob/master/examples/twisted/wamp/pubsub/simple/example2/client.py
import sys
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from twisted.internet.defer import inlineCallbacks
from twisted.python import log
from twisted.internet import reactor
from autobahn.twisted.websocket import connectWS, WampWebSocketClientProtocol, WampWebSocketClientFactory
from kotori.util.common import NodeId, get_hostname
from kotori.io.node.util import tts_say

node_manager = None
NODE_ID = str(NodeId())
NODE_HOSTNAME = get_hostname()

class NodeProtocol(WampWebSocketClientProtocol):
    """
    Implements various node services:
    - Heartbeat
    - Text-to-Speech
    """

    def heartbeat(self):
        self.publish("broadcast:node-heartbeat", NODE_ID)
        # FIXME: make this configurable as "heartbeat interval"
        reactor.callLater(60, self.heartbeat)

    def say(self, topic, message, language='de'):
        return tts_say(message, language=language)

    def onSessionOpen(self):

        self.prefix("broadcast", "de.elmyra.kotori.broadcast")
        self.prefix("presence", "http://kotori.elmyra.de/presence?")
        self.prefix("node", "http://kotori.elmyra.de/node/{0}#".format(NODE_ID))

        self.subscribe("broadcast:node-heartbeat", self.dump_event)

        self.subscribe("presence:node_id={0}&hostname={1}".format(NODE_ID, NODE_HOSTNAME), lambda: x)

        self.subscribe("node:say", self.dump_event)
        self.subscribe("node:say", self.say)

        print "INFO:   -> Node successfully connected to master"

        self.heartbeat()

        reactor.callLater(0, node_manager.start_features, self)


    def connectionLost(self, reason):
        print "ERROR: Connection lost, reason:", reason
        #reactor.callLater(2, node_manager.connect)

    def dump_event(self, topic, event):
        print "dump_event:", topic, event


class NodeClientFactory(WampWebSocketClientFactory):

    protocol = NodeProtocol
    noisy = False

    def __init__(self, url, **kwargs):
        return WampWebSocketClientFactory.__init__(self, url, **kwargs)

    def clientConnectionFailed(self, connector, reason):
        print 'NodeClientFactory.clientConnectionFailed:', reason

    def clientConnectionLost(self, connector, reason):
        print 'NodeClientFactory.clientConnectionLost:', reason

    def stopFactory(self):
        print 'NodeClientFactory.stopFactory'
        WampClientFactory.stopFactory(self)
        # FIXME: make this configurable as "reconnect interval"
        reactor.callLater(2, node_manager.master_connect)


class NodeManager(object):

    def __init__(self, websocket_uri, debug=False):
        self.websocket_uri = websocket_uri
        self.debug = debug

    def master_connect(self):
        factory = NodeClientFactory(self.websocket_uri, debug=False, debugCodePaths=False, debugWamp=self.debug, debugApp=False)
        connectWS(factory)

    def start_features(self, protocol):

        from kotori.io.node.feature import FeatureSet
        features = FeatureSet(NODE_ID, protocol)

        #actor = GpioOutput('P8_13')
        #actor.blink(0.2)


class KotoriNode(ApplicationSession):

    """
    An application component using the time service
    during 3 subsequent WAMP sessions, while the
    underlying transport continues to exist.
    """

    def __init__(self, config):
        ApplicationSession.__init__(self, config)
        self.count = 0

    @inlineCallbacks
    def onJoin(self, details):
        print("Realm joined (WAMP session started).")

        #self.heartbeat_loop()

        self.subscribe(self.heartbeat, u'de.elmyra.kotori.broadcast.heartbeat')
        self.subscribe(self.dump_event, u'de.elmyra.kotori.broadcast.operator-presence')

        try:
            now = yield self.call('com.timeservice.now')
        except Exception as e:
            print("Error: {}".format(e))
        else:
            print("Current time from time service: {}".format(now))


        yield self.publish(u'de.elmyra.kotori.telemetry.data', 'hello world')

        #self.leave()

    def onLeave(self, details):
        print("Realm left (WAMP session ended).")
        self.count += 1
        if self.count < 3:
            self.join("kotori-realm")
        else:
            self.disconnect()
        ApplicationSession.onLeave(self, details)

    def onDisconnect(self):
        print("Transport disconnected.")
        #reactor.stop()

    def heartbeat_loop(self, *args, **kwargs):
        self.heartbeat(*args, **kwargs)
        # FIXME: make this configurable as "heartbeat interval"
        reactor.callLater(30, self.heartbeat_loop)

    def heartbeat(self, *args, **kwargs):
        print("KotoriNode.heartbeat publish")
        self.publish(u'de.elmyra.kotori.broadcast.heartbeat', NODE_ID)


    def dump_event(self, *args, **kwargs):
        print "dump_event:", {'args': args, 'kwargs': kwargs}
        #print self, dir(self)


def boot_node(websocket_uri, debug=False, trace=False):

    print 'INFO: Starting node service, connecting to', websocket_uri

    # connect to master service
    """
    global node_manager
    node_manager = NodeManager(websocket_uri, debug=debug)
    reactor.callLater(0, node_manager.master_connect)
    """

    runner = ApplicationRunner(websocket_uri, u'kotori-realm', debug=trace, debug_wamp=debug, debug_app=debug)
    runner.run(KotoriNode, start_reactor=False)

    #app = Application()
    #app.run(url=websocket_uri, realm='kotori-realm', debug=True, debug_wamp=True, debug_app=True, start_reactor=False)


def run():
    log.startLogging(sys.stdout)

    if len(sys.argv) > 1 and sys.argv[1] == 'debug':
        debug = True
    else:
        debug = False

    # startup greeting
    #tts_say('Herzlich Willkommen')

    WEBSOCKET_URI = 'ws://localhost:9000'
    #WEBSOCKET_URI = 'ws://master.example.com:9000'
    boot_node(WEBSOCKET_URI, debug)

    reactor.run()

if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
# (c) 2017 Andreas Motl <andreas@getkotori.org>
import json

from pkg_resources import resource_filename
from jinja2 import Template
from twisted.logger import Logger
from grafana_api_client import GrafanaPreconditionFailedError, GrafanaClientError
from kotori.daq.services.mig import MqttInfluxGrafanaService
from kotori.daq.graphing.grafana.manager import GrafanaManager
from kotori.daq.storage.influx import InfluxDBAdapter

log = Logger()

class LuftdatenGrafanaManager(GrafanaManager):

    def __init__(self, *args, **kwargs):
        GrafanaManager.__init__(self, *args, **kwargs)
        self.tpl_dashboard_map      = self.get_template('grafana-map.json')
        self.tpl_dashboard_location = self.get_template('grafana-by-location.json')

    def get_template(self, filename):
        return Template(file(resource_filename('kotori.vendor.luftdaten', filename)).read().decode('utf-8'))

    def provision(self, storage_location, message, topology):

        topology = topology or {}
        dashboard_name = self.strategy.topology_to_label(topology)

        # The identity information of this provisioning process
        signature = (storage_location.database, storage_location.measurement)
        whoami = 'dashboard "{dashboard_name}" for database "{database}" and measurement "{measurement}"'.format(
            dashboard_name=dashboard_name, database=storage_location.database, measurement=storage_location.measurement)

        # Skip dashboard creation if it already has been created while Kotori is running
        # TODO: Improve locking to prevent race conditions.
        if self.keycache.exists(*signature):
            log.debug('Data signature not changed, skip update of {whoami}', whoami=whoami)
            return

        log.info('Provisioning Grafana {whoami}', whoami=whoami)

        # Create a Grafana datasource object for designated database
        self.create_datasource(storage_location)


        # Create appropriate Grafana dashboard

        data_dashboard = {
            'database': storage_location.database,
            'measurement': storage_location.measurement,
            'measurement_events': storage_location.measurement_events,
        }
        dashboard_json_map      = self.tpl_dashboard_map.render(data_dashboard, title='{name} map'.format(name=dashboard_name))
        dashboard_json_location = self.tpl_dashboard_location.render(data_dashboard, title='{name} by-location'.format(name=dashboard_name))

        # Get or create Grafana folder for stuffing all instant dashboards into
        folder = self.grafana_api.ensure_instant_folder()
        folder_id = folder and folder.get('id') or None

        for dashboard_json in [dashboard_json_map, dashboard_json_location]:

            try:
                log.info('Creating/updating dashboard "{}"'.format(dashboard_name))
                response = self.grafana_api.grafana_client.dashboards.db.create(
                    folderId=folder_id, dashboard=json.loads(dashboard_json), overwrite=True)
                log.info(u'Grafana response: {response}', response=json.dumps(response))

            except GrafanaPreconditionFailedError as ex:
                if 'name-exists' in ex.message or 'A dashboard with the same name already exists' in ex.message:
                    log.warn(ex.message)
                else:
                    log.error('Grafana Error: {ex}', ex=ex.message)

            except GrafanaClientError as ex:
                log.error('Grafana Error: {ex}', ex=ex.message)

        # Remember dashboard/panel creation for this kind of data inflow
        self.keycache.set(storage_location.database, storage_location.measurement)


class LuftdatenMqttInfluxGrafanaService(MqttInfluxGrafanaService):

    def setupService(self):
        MqttInfluxGrafanaService.setupService(self)
        self.settings.influxdb.use_udp = True
        self.settings.influxdb.udp_port = 4445
        self.influx = InfluxDBAdapter(settings = self.settings.influxdb)



#!/usr/bin/python
import os
import time
import Adafruit_BBIO.GPIO as GPIO

def hwports():
    for portno in range(0, 25):
        port = 'P8_' + str(portno)
        yield port

for port in hwports():
    GPIO.setup(port, GPIO.IN)

while True:
    os.system('clear')
    for port in hwports():
        value = GPIO.input(port)
        print port, value
    time.sleep(0.2)



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl, <andreas@getkotori.org>
import arrow
from twisted.logger import Logger
from twisted.application.service import MultiService

from kotori.daq.services import MultiServiceMixin
from kotori.daq.graphing.grafana.api import GrafanaApi
from kotori.daq.graphing.grafana.dashboard import GrafanaDashboardBuilder, GrafanaDashboardModel
from kotori.daq.graphing.grafana.service import DashboardRefreshTamingService
from kotori.util.common import KeyCache, SmartBunch

log = Logger()


class GrafanaManager(MultiService, MultiServiceMixin):

    def __init__(self, settings=None, channel=None):

        MultiService.__init__(self)

        # Make channel object from application settings configuration object
        self.setupChannel(channel=channel)

        # Shortcut to global settings
        self.config = settings

        if not 'port' in self.config['grafana']:
            self.config['grafana']['port'] = '3000'

        name = self.__class__.__name__
        log.info('Starting GrafanaManager "{}". grafana={}:{}'.format(
            name,
            self.config['grafana']['host'],
            self.config['grafana']['port']))

        # Initialize key cache
        # Utility functions for remembering whether the dashboard has been created already.
        # This is important as we would otherwise talk to Grafana for each ingress measurement (on each hit).
        self.keycache = KeyCache()

        # Boot further child services
        self.boot_workers()

        # Connect to Grafana API
        self.connect()

    def boot_workers(self):

        # Optionally enable dashboard refresh interval taming as yet another child service
        taming = self.channel.get('dashboard_refresh_taming')

        if taming:

            # Configuration convenience
            if taming == 'true':
                taming = 'standard'

            # Boot and register worker component
            service = DashboardRefreshTamingService(channel=self.channel, preset=taming)
            self.registerService(service)

    def connect(self):

        # TODO: Improve multi-tenancy / per-user isolation by using distinct configuration values and credentials per ingress channel.

        self.grafana_api = GrafanaApi(
            host = self.config['grafana']['host'],
            port = int(self.config['grafana']['port']),
            username = self.config['grafana']['username'],
            password = self.config['grafana']['password'],
        )

    def create_datasource(self, storage_location):

        datasource_name = storage_location.database

        self.grafana_api.create_datasource(datasource_name, {
            "type":     "influxdb",
            "url":      "http://{host}:{port}/".format(
                host=self.config['influxdb']['host'],
                port=int(self.config['influxdb'].get('port', '8086'))),
            "database": storage_location.database,
            "user":     self.config['influxdb']['username'],
            "password": self.config['influxdb']['password'],
            })

        return datasource_name

    def get_dashboard_identity(self, storage_location, topology=None):

        # Compute effective topology information
        topology = topology or {}
        realm = topology.get('realm', 'default')
        network = topology.get('network', storage_location.database)

        # Derive dashboard uid and name from topology information
        identity = SmartBunch(
            #uid=u'{realm}-{network}-instant'.format(realm=realm, network=network),
            name=u'{realm}-{network}'.format(realm=realm, network=network),
            title=u'{realm}-{network}'.format(realm=realm, network=network),
            # TODO: Use real title after fully upgrading to new Grafana API (i.e. don't use get-by-slug anymore!)
            #title=u'Raw data for realm={realm} network={network}'.format(realm=realm, network=network),
        )
        #print identity.prettify()

        return identity

    def provision(self, storage_location, data, topology=None):

        # TODO: Get into templating, finally: Create a template variable for each InfluxDB tag
        # TODO: Also provision a WorldMap plugin

        # The identity information of this provisioning process
        dashboard_identity = self.get_dashboard_identity(storage_location, topology)
        signature = (storage_location.database, storage_location.gateway, storage_location.node, data)
        whoami = u'dashboard "{dashboard_name}" for database "{database}" and measurement "{measurement}"'.format(
            dashboard_name=dashboard_identity.name, database=storage_location.database, measurement=storage_location.measurement)

        # Skip dashboard creation if it already has been created while Kotori is running
        # TODO: Improve locking to prevent race conditions.
        if self.keycache.exists(*signature):
            log.debug(u'Data signature not changed, skip update of {whoami}', whoami=whoami)
            return

        log.info(u'Provisioning Grafana {whoami}', whoami=whoami)

        # Create a Grafana datasource object for designated database
        datasource_name = self.create_datasource(storage_location)

        # Define Grafana dashboard model
        model = GrafanaDashboardModel(
            #uid=dashboard_identity.uid,
            name=dashboard_identity.name,
            title=dashboard_identity.title,
            datasource=datasource_name,
            measurement_sensors=storage_location.measurement,
            measurement_events=storage_location.measurement_events
        )

        # Create appropriate Grafana dashboard
        dashboard_builder = GrafanaDashboardBuilder(
            grafana_api=self.grafana_api,
            channel=self.channel,
            topology=topology,
            model=model
        )
        dashboard_builder.make(data=data)

        # Remember dashboard/panel creation for this kind of data inflow
        self.keycache.set(*signature)

    def tame_refresh_interval(self, preset='standard', force=False):
        """
        Tame refresh interval for all dashboards.

        :param mode: Which taming preset to use. Currently, only "standard" is
                     implemented, which is also the default preset.

        Introduction
        ------------
        The default dashboard refresh interval of 5 seconds is important
        for instant-on workbench operations. However, the update interval
        is usually just about 5 minutes after the sensor node is in the field.

        Problem
        -------
        Having high refresh rates on many dashboards can increase the overall
        system usage significantly, depending on how many users are displaying
        them in their browsers and the complexity of the database queries
        issued when rendering the dashboard.

        Solution
        --------
        In order to reduce the overall load on the data acquisition system,
        the refresh interval of dashboards not updated since a configurable
        threshold time is decreased according to rules of built-in presets.

        The default "standard" preset currently implements the following rules:

        - Leave all dashboards completely untouched which have been updated during the last 14 days
        - Apply a refresh interval of 5 minutes for all dashboards having the "live" tag
        - Completely disable refreshing for all dashboards having the "historical" tag
        - Apply a refresh interval of 30 minutes for all other dashboards

        """

        dashboard_list = self.grafana_api.get_dashboards()

        log.info('Taming dashboard refresh interval with preset="{preset}" for {count} dashboards',
                 preset=preset, count=len(dashboard_list))

        # Date of 14 days in the past
        before_14_days = arrow.utcnow().shift(days=-14)

        for dashboard_meta in dashboard_list:

            dashboard_meta = SmartBunch.bunchify(dashboard_meta)
            #print dashboard_meta.prettify()

            whoami = u'title="{title}", uid="{uid}"'.format(title=dashboard_meta['title'], uid=dashboard_meta['uid'])

            # Request dashboard by uid
            dashboard_uid = dashboard_meta['uid']
            response = self.grafana_api.get_dashboard_by_uid(dashboard_uid)
            response = SmartBunch.bunchify(response)

            # Get effective dashboard information from response
            folder_id = response.meta.folderId
            dashboard = response.dashboard

            # Compute new dashboard refresh interval by applying taming rules
            # Units: Mwdhmsy

            # 1. Check dashboard modification time against threshold
            modification_time = arrow.get(response.meta.updated)
            if not force and modification_time > before_14_days:
                log.debug('Skip taming dashboard with {whoami}, it has recently been modified', whoami=whoami)
                continue

            # 2. Apply refresh interval by looking at the dashboard tags
            if 'live' in dashboard_meta.tags:
                refresh_interval = '5m'
            elif 'historical' in dashboard_meta.tags:
                refresh_interval = None
            else:
                refresh_interval = '30m'

            # Skip update procedure if refresh interval hasn't changed at all
            if refresh_interval == dashboard.refresh:
                continue

            # Set new refresh interval
            dashboard.refresh = refresh_interval

            # Update dashboard
            log.debug('Taming dashboard with {whoami} to refresh interval of {interval}', whoami=whoami, interval=refresh_interval)
            response = self.grafana_api.grafana_client.dashboards.db.create(dashboard=dashboard, folderId=folder_id)

            # Report about the outcome
            if response['status'] == 'success':
                log.info('Successfully tamed dashboard with {whoami}', whoami=whoami)
            else:
                log.warn('Failed taming dashboard with {whoami}', whoami=whoami)


if __name__ == '__main__':
    """
    Example usage of GrafanaApi and GrafanaDashboard objects.
    """

    # Bootstrap logging
    import sys
    import twisted
    from kotori.daq.graphing.grafana.dashboard import GrafanaDashboard

    twisted.python.log.startLogging(sys.stderr)

    # Connect to Grafana
    manager = GrafanaManager(
        settings={"grafana": dict(host='localhost', username='admin', password='admin')},
        channel={}
    )
    grafana = manager.grafana_api

    # Create Grafana Datasource object
    grafana.create_datasource('hiveeyes_test', {
        "type":     "influxdb",
        "url":      "http://localhost:8086/",
        "database": "hiveeyes_test",
        "user":     "root",
        "password": "root",
    })

    # Get or create Grafana folder for stuffing all instant dashboards into
    folder = grafana.ensure_instant_folder()
    folder_id = folder and folder.get('id') or None

    # Create dashboard
    dashboard = GrafanaDashboard(datasource='hiveeyes_test', title='hiveeyes_test', folder_id=folder_id)
    dashboard.build(measurement='1_2', row_title='node=2,gw=1', panels=[
        {'title': 'temp',      'fieldnames': ['temp1', 'temp2', 'temp3', 'temp4'], 'format': 'celsius'},
        {'title': 'humidity',  'fieldnames': ['hum1', 'hum2']},
        {'title': 'weight',    'fieldnames': ['wght1'], 'label': 'kg'},
    ])
    grafana.create_dashboard(dashboard)

    # Run one-shot task to tame the dashboard intervals
    manager.tame_refresh_interval()



# -*- coding: utf-8 -*-
# (c) 2015-2018 Andreas Motl <andreas@getkotori.org>
import types
import requests
import datetime
from sets import Set
from copy import deepcopy
from funcy import project
from collections import OrderedDict
from twisted.logger import Logger
from influxdb.client import InfluxDBClient, InfluxDBClientError
from kotori.io.protocol.util import parse_timestamp, is_number, convert_floats

log = Logger()


class InfluxDBAdapter(object):

    def __init__(self, settings=None, database=None):

        settings = deepcopy(settings) or {}
        settings.setdefault('host', u'localhost')
        settings.setdefault('port', u'8086')
        settings.setdefault('username', u'root')
        settings.setdefault('password', u'root')
        settings.setdefault('database', database)

        settings.setdefault('use_udp', False)
        settings.setdefault('udp_port', u'4444')

        settings['port'] = int(settings['port'])
        settings['udp_port'] = int(settings['udp_port'])

        self.__dict__.update(**settings)

        # Bookeeping for all databases having been touched already
        self.databases_written_once = Set()

        # Knowledge about all databases to be accessed using UDP
        # TODO: Refactor to configuration setting
        self.udp_databases = [
            {'name': 'luftdaten_info', 'port': u'4445'},
        ]
        self.host_uri = u'influxdb://{host}:{port}'.format(**self.__dict__)

        log.info(u'Storage target is {uri}', uri=self.host_uri)
        self.influx_client = InfluxDBClient(
            host=self.host, port=self.port,
            username=self.username, password=self.password,
            database=self.database,
            timeout=10)

        # TODO: Hold references to multiple UDP databases using mapping "self.udp_databases".
        self.influx_client_udp = None
        if settings['use_udp']:
            self.influx_client_udp = InfluxDBClient(
                host=self.host, port=self.port,
                username=self.username, password=self.password,
                use_udp=settings['use_udp'], udp_port=settings['udp_port'],
                timeout=10)

    def is_udp_database(self, name):
        for entry in self.udp_databases:
            if entry['name'] == name:
                return True
        return False

    def write(self, meta, data):

        meta_copy = deepcopy(dict(meta))
        data_copy = deepcopy(data)

        try:
            chunk = self.format_chunk(meta, data)

        except Exception as ex:
            log.failure(u'Could not format chunk (ex={ex_name}: {ex}): data={data}, meta={meta}',
                ex_name=ex.__class__.__name__, ex=ex, meta=meta_copy, data=data_copy)
            raise

        try:
            success = self.write_chunk(meta, chunk)
            return success

        except requests.exceptions.ConnectionError as ex:
            log.failure(u'Problem connecting to InfluxDB at {uri}: {ex}', uri=self.host_uri, ex=ex)
            raise

        except InfluxDBClientError as ex:

            if ex.code == 404 or ex.message == 'database not found':

                log.info('Creating database "{database}"', database=meta.database)
                self.influx_client.create_database(meta.database)

                # Attempt second write
                success = self.write_chunk(meta, chunk)
                return success

                #log.failure('InfluxDBClientError: {ex}', ex=ex)

            # [0.8] ignore "409: database kotori-dev exists"
            # [0.9] ignore "database already exists"
            elif ex.code == 409 or ex.message == 'database already exists':
                pass
            else:
                raise

    def write_chunk(self, meta, chunk):
        if self.influx_client_udp and self.is_udp_database(meta.database) and meta.database in self.databases_written_once:
            success = self.influx_client_udp.write_points([chunk], time_precision='s', database=meta.database)
        else:
            success = self.influx_client.write_points([chunk], time_precision=chunk['time_precision'], database=meta.database)
            self.databases_written_once.add(meta.database)
        if success:
            log.debug(u"Storage success: {chunk}", chunk=chunk)
        else:
            log.error(u"Storage failed:  {chunk}", chunk=chunk)
        return success

    @staticmethod
    def get_tags(data):
        return project(data, ['gateway', 'node'])

    def format_chunk(self, meta, data):
        """
        Format for InfluxDB >= 0.9::
        {
            "measurement": "hiveeyes_100",
            "tags": {
                "host": "server01",
                "region": "europe"
            },
            "time": "2015-10-17T19:30:00Z",
            "fields": {
                "value": 0.42
            }
        }
        """

        assert isinstance(data, dict), 'Data payload is not a dictionary'

        chunk = {
            "measurement": meta['measurement'],
            "tags": {},
        }

        """
        if "gateway" in meta:
            chunk["tags"]["gateway"] = meta["gateway"]

        if "node" in meta:
            chunk["tags"]["node"]    = meta["node"]
        """

        # Extract timestamp field from data
        chunk['time_precision'] = 'n'
        for time_field in ['time', 'datetime', 'dateTime']:
            if time_field in data:

                # WeeWX. TODO: Move to specific vendor configuration.
                # Disabled in favor of precision detection heuristic.
                #if time_field == 'dateTime':
                #    chunk['time_precision'] = 's'

                # Process timestamp field.
                if data[time_field]:

                    # Decode timestamp.
                    chunk['time'] = data[time_field]
                    if is_number(chunk['time']):
                        chunk['time'] = int(float(chunk['time']))

                    # Remove timestamp from data payload.
                    del data[time_field]

                    # If we found a timestamp field already,
                    # don't look out for more.
                    break

        # Extract geohash from data. Finally, thanks Rich!
        # TODO: Also precompute geohash with 3-4 different zoomlevels and add them as tags
        if "geohash" in data:
            chunk["tags"]["geohash"] = data["geohash"]
            del data['geohash']

        if "latitude" in data and "longitude" in data:
            chunk["tags"]["latitude"] = data["latitude"]
            chunk["tags"]["longitude"] = data["longitude"]
            del data['latitude']
            del data['longitude']

        # Extract more information specific to luftdaten.info
        for field in ['location', 'location_id', 'location_name', 'sensor_id', 'sensor_type']:
            if field in data:
                chunk["tags"][field] = data[field]
                del data[field]

        # TODO: Maybe do this at data acquisition / transformation time, not here.
        if 'time' in chunk:
            timestamp = chunk['time'] = parse_timestamp(chunk['time'])

            # Heuristically compute timestamp precision
            if isinstance(timestamp, int):
                if timestamp >= 1e17 or timestamp <= -1e17:
                    time_precision = 'n'
                elif timestamp >= 1e14 or timestamp <= -1e14:
                    time_precision = 'u'
                elif timestamp >= 1e11 or timestamp <= -1e11:
                    time_precision = 'ms'

                # FIXME: Is this a reasonable default?
                else:
                    time_precision = 's'

                chunk['time_precision'] = time_precision

            """
            # FIXME: Breaks CSV data acquisition. Why?
            if isinstance(chunk['time'], datetime.datetime):
                if chunk['time'].microsecond == 0:
                    chunk['time_precision'] = 's'
            """

        """
        Prevent errors like
        ERROR: InfluxDBClientError: 400:
                       write failed: field type conflict:
                       input field "pitch" on measurement "01_position" is type float64, already exists as type integer
        """
        self.data_to_float(data)

        assert data, 'Data payload is empty'

        chunk["fields"] = data

        return chunk

    def data_to_float(self, data):
        return convert_floats(data)

        for key, value in data.iteritems():

            # Sanity checks
            if type(value) in types.StringTypes:
                continue

            if value is None:
                data[key] = None
                continue

            # Convert to float
            try:
                data[key] = float(value)
            except (TypeError, ValueError) as ex:
                log.warn(u'Measurement "{key}: {value}" float conversion failed: {ex}', key=key, value=value, ex=ex)


class BusInfluxForwarder(object):
    """
    Generic software bus -> influxdb forwarder based on prototypic implementation at HiveEyes
    TODO: Generalize and refactor
    """

    # TODO: Improve parameter passing
    def __init__(self, bus, topic, config, channel):
        self.bus = bus
        self.topic = topic
        self.config = config
        self.channel = channel

        self.bus.subscribe(self.bus_receive, self.topic)


    def topic_to_topology(self, topic):
        raise NotImplementedError()

    @staticmethod
    def sanitize_db_identifier(value):
        value = unicode(value).replace('/', '_').replace('.', '_').replace('-', '_')
        return value

    def bus_receive(self, payload):
        try:
            return self.process_message(self.topic, payload)
        except Exception:
            log.failure(u'Processing bus message failed')

    def process_message(self, topic, payload, *args):

        log.debug('Bus receive: topic={topic}, payload={payload}', topic=topic, payload=payload)

        # TODO: filter by realm/topic

        # decode message
        if type(payload) is types.DictionaryType:
            message = payload.copy()
        elif type(payload) is types.ListType:
            message = OrderedDict(payload)
        else:
            raise TypeError('Unable to handle data type "{}" from bus'.format(type(payload)))

        # compute storage location from topic and message
        storage_location = self.storage_location(message)
        log.debug('Storage location: {storage_location}', storage_location=dict(storage_location))

        # store data
        self.store_message(storage_location, message)


    def storage_location(self, data):
        raise NotImplementedError()

    def store_encode(self, data):
        return data

    def store_message(self, location, data):

        data = self.store_encode(data)

        influx = InfluxDBAdapter(
            settings = self.config['influxdb'],
            database = location.database)

        outcome = influx.write(location, data)
        log.debug('Store outcome: {outcome}', outcome=outcome)

        self.on_store(location, data)

    def on_store(self, location, data):
        pass




# -*- coding: utf-8 -*-
"""Implementation of a Twisted-friendly thread pool wrapper."""
from __future__ import absolute_import
from functools import partial
from twisted.internet.threads import deferToThreadPool
from twisted.internet.defer import fail
from twisted.internet.error import ReactorNotRunning

# Shamelessly stolen from:
# https://github.com/lvh/thimble
# https://pypi.python.org/pypi/thimble

class Thimble(object):

    """A Twisted thread-pool wrapper for a blocking API."""

    def __init__(self, reactor, pool, wrapped, blocking_methods):
        """Initialize a :class:`Thimble`.
        :param reactor: The reactor that will handle events.
        :type reactor: :class:`twisted.internet.interfaces.IReactorThreads` and
            :class:`twisted.internet.interfaces.IReactorCore`. Pretty much any
            real reactor implementation will do.
        :param pool: The thread pool to defer to.
        :type pool: :class:`twisted.python.threadpool.ThreadPool`
        :param wrapped: The blocking implementation being wrapped.
        :param blocking_methods: The names of the methods that will be wrapped
            and executed in the thread pool.
        :type blocking_methods: ``list`` of native ``str``
        """
        self._reactor = reactor
        self._pool = pool
        self._wrapped = wrapped
        self._blocking_methods = blocking_methods

    def _deferToThreadPool(self, f, *args, **kwargs):
        """Defer execution of ``f(*args, **kwargs)`` to the thread pool.
        This returns a deferred which will callback with the result of
        that expression, or errback with a failure wrapping the raised
        exception.
        """
        if self._pool.joined:
            return fail(
                ReactorNotRunning("This thimble's threadpool already stopped.")
            )
        if not self._pool.started:
            self._pool.start()
            self._reactor.addSystemEventTrigger(
                'during', 'shutdown', self._pool.stop)

        return deferToThreadPool(self._reactor, self._pool, f, *args, **kwargs)

    def __getattr__(self, attr):
        """Get and maybe wraps an attribute from the wrapped object.
        If the attribute is blocking, it will be wrapped so that
        calling it will return a Deferred and the actual function will
        be ran in a thread pool.
        """
        value = getattr(self._wrapped, attr)

        if attr in self._blocking_methods:
            value = partial(self._deferToThreadPool, value)

        return value



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

#from kotori.daq.intercom.mqtt.twisted import TwistedMqttAdapter
#MqttAdapter = TwistedMqttAdapter

from kotori.daq.intercom.mqtt.paho import PahoMqttAdapter
MqttAdapter = PahoMqttAdapter



# -*- coding: utf-8 -*-
# (c) 2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
from sqlalchemy.engine import create_engine
from sqlalchemy.sql.ddl import CreateTable
from sqlalchemy.sql.schema import MetaData, Column, Table
from sqlalchemy.sql.sqltypes import Integer, String, Numeric
from alchimia.strategy import TWISTED_STRATEGY
from autobahn.twisted.wamp import ApplicationRunner, ApplicationSession
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

class SqlDatabaseService(ApplicationSession):
    """An application component for logging telemetry data to sql databases"""

    def __init__(self, config):
        ApplicationSession.__init__(self, config)
        self.count = 0
        self.engine = None


        metadata = MetaData()
        self.telemetry = Table("telemetry", metadata,
            Column("id", Integer(), primary_key=True),
            Column("MSG_ID", Integer()),
            Column("V_FC", Integer()),
            Column("V_CAP", Integer()),
            Column("A_ENG", Integer()),
            Column("A_CAP", Integer()),
            Column("T_O2_In", Integer()),
            Column("T_O2_Out", Integer()),
            Column("T_FC_H2O_Out", Integer()),
            Column("Water_In", Integer()),
            Column("Water_Out", Integer()),
            Column("Master_SW", Integer()),
            Column("CAP_Down_SW", Integer()),
            Column("Drive_SW", Integer()),
            Column("FC_state", Integer()),
            Column("Mosfet_state", Integer()),
            Column("Safety_state", Integer()),
            Column("Air_Pump_load", Numeric()),
            Column("Mosfet_load", Integer()),
            Column("Water_Pump_load", Integer()),
            Column("Fan_load", Integer()),
            Column("Acc_X", Integer()),
            Column("Acc_Y", Integer()),
            Column("Acc_Z", Integer()),
            Column("AUX", Numeric()),
            Column("GPS_X", Integer()),
            Column("GPS_Y", Integer()),
            Column("GPS_Z", Integer()),
            Column("GPS_Speed", Integer()),
            Column("V_Safety", Integer()),
            Column("H2_Level", Integer()),
            Column("O2_calc", Numeric()),
            Column("lat", Numeric()),
            Column("lng", Numeric()),
            )


#        metadata = MetaData()
#        self.telemetry = Table("telemetry", metadata,
#            Column("id", Integer(), primary_key=True),
#            Column("mma_x", Integer()),
#            Column("mma_y", Integer()),
#            Column("temp", Numeric()),
#	    Column("lat", Numeric()),
#	    Column("lng", Numeric()),
#        )

    #@inlineCallbacks
    def onJoin(self, details):
        print("Realm joined (WAMP session started).")

        # subscribe to telemetry data channel
        self.subscribe(self.receive, u'de.elmyra.kotori.telemetry.data')

        self.startDatabase()

        #self.leave()

    #@inlineCallbacks
    def startDatabase(self):
        self.engine = create_engine(

            # sqlite in-memory
            #"sqlite://", reactor=reactor, strategy=TWISTED_STRATEGY

            # sqlite on filesystem
            "sqlite:////tmp/kotori.sqlite", reactor=reactor, strategy=TWISTED_STRATEGY

            # mysql... todo
        )

        # Create the table
        yield self.engine.execute(CreateTable(self.telemetry))
        #yield self.engine

    def onLeave(self, details):
        print("Realm left (WAMP session ended).")
        ApplicationSession.onLeave(self, details)

    def onDisconnect(self):
        print("Transport disconnected.")
        #reactor.stop()

    #@inlineCallbacks
    def receive(self, data):
        #print "RECEIVE:", data

        # decode wire data
        payload = data.split(';')
        MSG_ID          = int(payload[0])
        V_FC            = int(payload[1])
        V_CAP           = int(payload[2])
        A_ENG           = int(payload[3])
        A_CAP           = int(payload[4])
        T_O2_In         = int(payload[5])
        T_O2_Out        = int(payload[6])
        T_FC_H2O_Out    = int(payload[7])
        Water_In        = int(payload[8])
        Water_Out       = int(payload[9])
        Master_SW       = int(payload[10])
        CAP_Down_SW     = int(payload[11])
        Drive_SW        = int(payload[12])
        FC_state        = int(payload[13])
        Mosfet_state    = int(payload[14])
        Safety_state    = int(payload[15])
        Air_Pump_load   = float(payload[16])
        Mosfet_load     = int(payload[17])
        Water_Pump_load = int(payload[18])
        Fan_load        = int(payload[19])
        Acc_X           = int(payload[20])
        Acc_Y           = int(payload[21])
        Acc_Z           = int(payload[22])
        AUX             = float(payload[23])
        GPS_X           = int(payload[24])
        GPS_Y           = int(payload[25])
        GPS_Z           = int(payload[26])
        GPS_Speed       = int(payload[27])
        V_Safety        = int(payload[28])
        H2_Level        = int(payload[29])
        O2_calc         = float(payload[30])
        lat             = float(payload[31])
        lng             = float(payload[32])




#        mma_x = int(payload[0])
#        mma_y = int(payload[1])
#        temp = float(payload[2])
#        try:
#            lat = float(payload[3])
#            lng = float(payload[4])
#        except:
#            lat = 0
#            lng = 0

        # store data to database
        if self.engine:
            yield self.engine.execute(self.telemetry.insert().values(MSG_ID = MSG_ID, V_FC = V_FC, V_CAP = V_CAP, A_ENG = A_ENG, A_CAP = A_CAP, T_O2_In = T_O2_In, T_O2_Out = T_O2_Out, T_FC_H2O_Out = T_FC_H2O_Out, Water_In = Water_In, Water_Out = Water_Out, Master_SW = Master_SW, CAP_Down_SW = CAP_Down_SW, Drive_SW = Drive_SW, FC_state = FC_state, Mosfet_state = Mosfet_state, Safety_state = Safety_state, Air_Pump_load = Air_Pump_load, Mosfet_load = Mosfet_load, Water_Pump_load = Water_Pump_load, Fan_load = Fan_load, Acc_X = Acc_X, Acc_Y = Acc_Y, Acc_Z = Acc_Z, AUX = AUX, GPS_X = GPS_X, GPS_Y = GPS_Y, GPS_Z = GPS_Z, GPS_Speed = GPS_Speed, V_Safety = V_Safety, H2_Level = H2_Level, lat = lat, lng = lng))


def boot_sql_database(websocket_uri, debug=False, trace=False):

    print 'INFO: Starting sql database service, connecting to broker', websocket_uri

    runner = ApplicationRunner(websocket_uri, u'kotori-realm', debug=trace, debug_wamp=debug, debug_app=debug)
    runner.run(SqlDatabaseService, start_reactor=False)



# -*- coding: utf-8 -*-
# (c) 2014-2015 Andreas Motl, Elmyra UG <andreas.motl@elmyra.de>
import sys
import time
import random
import socket
from twisted.python import log


# derived from https://twistedmatrix.com/documents/15.0.0/core/howto/udp.html
def run_udp_client():
    log.startLogging(sys.stdout)

    #data = 'UDP hello'
    #data = '5;3;'
    try:
        data = sys.argv[1]
    except IndexError:
        data = ''

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.sendto(data, ('127.0.0.1', 7777))



def generate_packet():
    packet_tpl = "24000;{};{};{};{};{};{};{};{};{};0;0;0;0;0;1;0;12;0;0;0;-18;0;4011;417633984;85402624;472851424;0;12242;43;42;0;0"
    values = []
    for i in range(20):
        value = random.randint(0, 99999)
        values.append(value)
    #print values
    packet = packet_tpl.format(*values)
    return packet

def run_udp_fuzzer():
    log.startLogging(sys.stdout)

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    while True:
        data = generate_packet()
        print 'packet:', data

        sock.sendto(data, ('127.0.0.1', 7777))
        time.sleep(0.25)



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import types
import pandas
import collections
from twisted.logger import Logger
from kotori.daq.storage.influx import InfluxDBAdapter

log = Logger()

class DataFrameQuery(object):
    """
    Query InfluxDB, massage result and return as DataFrame.
    """

    def __init__(self, settings=None, bucket=None):
        self.settings = settings
        self.bucket = bucket
        self.request = bucket.request

    def query(self):

        bucket = self.bucket

        # Get an InfluxDB adapter object
        # TODO: Pool these, keyed by database.
        influx = InfluxDBAdapter(
            settings = self.settings.influxdb,
            database = bucket.tdata.database)

        # Get query expression from transformation dictionary
        expression = bucket.tdata.expression

        log.info('InfluxDB expression: {expression}', expression=expression)

        # Run database query
        result = influx.influx_client.query(expression)

        # Massage result format
        entries = list(flatten(result))

        # Stop when having no results
        if not entries:
            return

        # Bring column names in order, "time" should be the first column
        columns = entries[0].keys()
        if 'time' in columns:
            columns.remove('time')
            columns.insert(0, 'time')

        # Make pandas DataFrame from database results
        df = pandas.DataFrame(entries, columns=columns)

        # Convert "time" column to datetime format
        df['time'] = pandas.to_datetime(df['time'])

        return df


def flatten(l):
    """
    Munge InfluxDB results.

    See also: https://stackoverflow.com/questions/21461140/flatten-an-irregular-list-of-lists-in-python-respecting-pandas-dataframes
    """
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el, (basestring, pandas.DataFrame, types.DictionaryType)):
            for sub in flatten(el):
                yield sub
        else:
            yield el



# -*- coding: utf-8 -*-
# (c) 2016-2017 Andreas Motl <andreas@getkotori.org>
import arrow
import pandas
from arrow.parser import DateTimeParser
from datetime import datetime, timedelta
from pyinfluxql import Query
from pyinfluxql.functions import Mean
from twisted.logger import Logger
from kotori.io.protocol.util import is_number
from kotori.util.common import tdelta

log = Logger()

class QueryTransformer(object):

    @classmethod
    def transform(cls, data):
        """
        Compute InfluxDB query expression from data in transformation dictionary.
        Also compute date range from query parameters "from" and "to".
        """

        measurement = data.measurement

        # Vanilla QL (v1)
        #expression = 'SELECT * FROM {measurement}'.format(measurement=measurement)

        # PyInfluxQL (v2)
        # https://github.com/jjmalina/pyinfluxql

        # Labs
        #time_begin = arrow.utcnow() - arrow.Arrow(hour=1)
        #expression = Query('*').from_(measurement).where(time__gt=datetime.utcnow() - timedelta(hours=1))
        #expression = Query(Mean('*')).from_(measurement).where(time__gt=datetime.now() - timedelta(1)).group_by(time=timedelta(hours=1))

        # Fix up "measurement" if starting with numeric value
        # TODO: Fix should go to pyinfluxql
        if is_number(measurement[0]):
            measurement = '"{measurement}"'.format(measurement=measurement)

        # TODO: Use ".date_range" API method
        time_begin, time_end = compute_daterange(data.get('from'), data.get('to'))

        tags = {}
        #tags = InfluxDBAdapter.get_tags(data)

        expression = Query('*').from_(measurement).where(time__gte=time_begin, time__lte=time_end, **tags)

        result = {
            'expression': str(expression),
            'time_begin': time_begin,
            'time_end':   time_end,
        }

        return result


def get_timedelta(expression):

    # TODO: Use pandas' Timedelta. Timedelta('1m2s')
    # http://pandas.pydata.org/pandas-docs/stable/timedeltas.html

    # FIXME: Sanitize expression
    code = expression
    delta_raw = code.replace('now-', '')
    if code != delta_raw:
        code = code.replace(delta_raw, 'delta')

    # "code" should now be "now-delta"
    #print 'code:', code

    now = datetime.utcnow()
    delta = tdelta(delta_raw)

    # FIXME: This is nasty
    try:
        td = eval(code)
    except:
        raise ValueError('Unknown expression: {expression}'.format(expression=expression))

    return td


def compute_daterange(raw_begin, raw_end):

    # Defaults
    raw_begin = raw_begin or 'now-10d'
    raw_end   = raw_end   or 'now'

    # Parse dates, absolute or relative
    time_begin = grok_datetime(raw_begin)
    time_end   = grok_datetime(raw_end)

    # If end of date range is supplied as date without time ('YYYY-MM-DD' or 'YYYYMMDD'),
    # add appropriate offset to mean "end of day" (DWIM).
    if 8 <= len(raw_end) <= 10:
        offset_endofday = tdelta('23h59m59s') + timedelta(microseconds = 999999)
        time_end += offset_endofday

    return time_begin, time_end

def grok_datetime(dstring):
    more_formats = ['YYYYMMDDTHHmmss', 'YYYYMMDDTHHmmssZ', 'YYYYMMDD']
    parser = DateTimeParser()
    parser.SEPARATORS += ['']

    # Try to parse datetime string in regular ISO 8601 format
    try:
        return parser.parse_iso(dstring)

    # Fall back to parse datetime string in additional convenience formats
    except arrow.parser.ParserError as ex:
        for format in more_formats:
            try:
                return parser.parse(dstring, format)
            except arrow.parser.ParserError as ex:
                pass

    # Fall back to attempt to parse as relative datetime expression, e.g. "now-10m"
    return get_timedelta(dstring)

